from __future__ import annotations

import copy
import grp
import os
import pwd
import re
import shutil
import stat
import struct
import sys
import tarfile
import time
from abc import ABC
from builtins import open as bltn_open
from typing import Any, Optional

from streamflow.core.data import StreamWrapper
from streamflow.data.stream import BaseStreamWrapper


async def copyfileobj(src, dst, length=None, bufsize=None):
    bufsize = bufsize or 16 * 1024
    if length == 0:
        return
    if length is None:
        shutil.copyfileobj(src, dst, bufsize)
        return
    blocks, remainder = divmod(length, bufsize)
    for b in range(blocks):
        await write(src, dst, bufsize)
    if remainder != 0:
        await write(src, dst, remainder)
    return


async def write(src, dst, bufsize):
    while bufsize > 0:
        buf = await src.read(bufsize) if isinstance(src, StreamWrapper) else src.read(bufsize)
        bufsize -= len(buf)
        await dst.write(buf) if isinstance(dst, StreamWrapper) else dst.write(buf)


class CompressionStreamWrapper(BaseStreamWrapper, ABC):

    def __init__(self, stream, mode):
        super().__init__(stream)
        self.mode = mode
        self.cmp = None
        self.exception = None

    async def close(self):
        if self.closed:
            return
        self.closed = True
        try:
            if self.mode == "w":
                await self.stream.write(self.cmp.flush())
        finally:
            await self.stream.close()

    async def read(self, size: Optional[int] = None):
        buf = await super().read(size)
        try:
            return self.cmp.decompress(buf)
        except self.exception:
            raise tarfile.ReadError("invalid compressed data")

    async def write(self, data: Any):
        data = self.cmp.compress(data)
        await super().write(data)


class BZ2StreamWrapper(CompressionStreamWrapper):

    def __init__(self, stream, mode, compresslevel: int = 9):
        super().__init__(stream, mode)
        try:
            import bz2
        except ImportError:
            raise tarfile.CompressionError("bz2 module is not available") from None
        if self.mode == "r":
            self.cmp = bz2.BZ2Decompressor()
            self.exception = OSError
        else:
            self.cmp = bz2.BZ2Compressor(compresslevel=compresslevel)


class GZipStreamWrapper(CompressionStreamWrapper):

    def __init__(self, stream, mode, compresslevel: int = 9):
        super().__init__(stream, mode)
        self.compresslevel: int = compresslevel
        self.cmp = None
        self.pos = 0
        try:
            import zlib
        except ImportError:
            raise tarfile.CompressionError("zlib module is not available") from None
        self.zlib = zlib
        self.crc = zlib.crc32(b"")
        if self.mode == "r":
            self.exception = zlib.error

    async def _init_compression(self):
        if self.mode == "r":
            await self._init_read_gz()
        else:
            await self._init_write_gz()

    async def _init_read_gz(self):
        self.cmp = self.zlib.decompressobj(-self.zlib.MAX_WBITS)
        if await self.stream.read(2) != b"\037\213":
            raise tarfile.ReadError("not a gzip file")
        if await self.stream.read(1) != b"\010":
            raise tarfile.CompressionError("unsupported compression method")
        flag = ord(await self.stream.read(1))
        await self.stream.read(6)
        if flag & 4:
            xlen = ord(await self.stream.read(1)) + 256 * ord(await self.stream.read(1))
            await self.stream.read(xlen)
        if flag & 8:
            while True:
                s = await self.stream.read(1)
                if not s or s == tarfile.NUL:
                    break
        if flag & 16:
            while True:
                s = await self.stream.read(1)
                if not s or s == tarfile.NUL:
                    break
        if flag & 2:
            await self.stream.read(2)

    async def _init_write_gz(self):
        self.cmp = self.zlib.compressobj(
            level=self.compresslevel,
            method=self.zlib.DEFLATED,
            wbits=-self.zlib.MAX_WBITS,
            memLevel=self.zlib.DEF_MEM_LEVEL,
            strategy=self.zlib.Z_DEFAULT_STRATEGY)
        timestamp = struct.pack("<L", int(time.time()))
        await self.stream.write(b"\037\213\010\010" + timestamp + b"\002\377" + tarfile.NUL)

    async def close(self):
        if self.closed:
            return
        self.closed = True
        try:
            await self.stream.write(self.cmp.flush())
            await self.stream.write(struct.pack("<L", self.crc))
            await self.stream.write(struct.pack("<L", self.pos & 0xffffFFFF))
        finally:
            await self.stream.close()

    async def read(self, size: Optional[int] = None):
        if self.cmp is None:
            await self._init_compression()
        buf = await super().read(size)
        self.pos += len(buf)
        return buf

    async def write(self, data: Any):
        if self.cmp is None:
            await self._init_compression()
        self.crc = self.zlib.crc32(data, self.crc)
        self.pos += len(data)
        await super().write(data)


class LZMAStreamWrapper(CompressionStreamWrapper):

    def __init__(self, stream, mode, preset: Optional[int] = None):
        super().__init__(stream, mode)
        try:
            import lzma
        except ImportError:
            raise tarfile.CompressionError("lzma module is not available") from None
        if self.mode == "r":
            self.cmp = lzma.LZMADecompressor()
            self.exception = lzma.LZMAError
        else:
            self.cmp = lzma.LZMACompressor(preset=preset)


class FileStreamReaderWrapper(StreamWrapper):

    def __init__(self, stream, size, blockinfo=None):
        super().__init__(stream)
        self.size = size
        self.position = 0
        self.closed = False
        if blockinfo is None:
            blockinfo = [(0, size)]
        self.map_index = 0
        self.map = []
        lastpos = 0
        for offset, size in blockinfo:
            if offset > lastpos:
                self.map.append((False, lastpos, offset))
            self.map.append((True, offset, offset + size))
            lastpos = offset + size
        if lastpos < self.size:
            self.map.append((False, lastpos, self.size))

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        self.closed = True

    async def read(self, size: Optional[int] = None):
        size = self.size - self.position if size is None else min(size, self.size - self.position)
        while True:
            data, start, stop = self.map[self.map_index]
            if start <= self.position < stop:
                break
            else:
                self.map_index += 1
                if self.map_index == len(self.map):
                    return bytes()
        length = min(size, stop - self.position)
        if data:
            buf = await self.stream.read(length)
            self.position += len(buf)
            return buf
        else:
            self.position += length
            return tarfile.NUL * length

    async def write(self, data: Any):
        raise NotImplementedError


class AioTarInfo(tarfile.TarInfo):

    @classmethod
    async def fromtarfile(cls, tarstream):
        buf = await tarstream.stream.read(tarfile.BLOCKSIZE)
        obj = cls.frombuf(buf, tarstream.encoding, tarstream.errors)
        obj.offset = tarstream.offset - tarfile.BLOCKSIZE
        return await obj._proc_member(tarstream)

    def _proc_builtin(self, tarstream):
        self.offset_data = tarstream.offset
        offset = self.offset_data
        if self.isreg() or self.type not in tarfile.SUPPORTED_TYPES:
            offset += self._block(self.size)
        tarstream.offset = offset
        self._apply_pax_info(tarstream.pax_headers, tarstream.encoding, tarstream.errors)
        return self

    async def _proc_gnulong(self, tarstream):
        buf = await tarstream.stream.read(self._block(self.size))
        try:
            next = await self.fromtarfile(tarstream)
        except tarfile.HeaderError as e:
            raise tarfile.SubsequentHeaderError(str(e)) from None
        next.offset = self.offset
        if self.type == tarfile.GNUTYPE_LONGNAME:
            next.name = tarfile.nts(buf, tarstream.encoding, tarstream.errors)
        elif self.type == tarfile.GNUTYPE_LONGLINK:
            next.linkname = tarfile.nts(buf, tarstream.encoding, tarstream.errors)
        return next

    async def _proc_gnusparse_10(self, next, tarstream):
        sparse = []
        buf = await tarstream.stream.read(tarfile.BLOCKSIZE)
        fields, buf = buf.split(b"\n", 1)
        fields = int(fields)
        while len(sparse) < fields * 2:
            if b"\n" not in buf:
                buf += await tarstream.stream.read(tarfile.BLOCKSIZE)
            number, buf = buf.split(b"\n", 1)
            sparse.append(int(number))
        next.offset_data = tarstream.offset
        next.sparse = list(zip(sparse[::2], sparse[1::2]))

    async def _proc_member(self, tarstream):
        if self.type in (tarfile.GNUTYPE_LONGNAME, tarfile.GNUTYPE_LONGLINK):
            return await self._proc_gnulong(tarstream)
        elif self.type == tarfile.GNUTYPE_SPARSE:
            return await self._proc_sparse(tarstream)
        elif self.type in (tarfile.XHDTYPE, tarfile.XGLTYPE, tarfile.SOLARIS_XHDTYPE):
            return await self._proc_pax(tarstream)
        else:
            return self._proc_builtin(tarstream)

    async def _proc_pax(self, tarstream):
        buf = await tarstream.stream.read(self._block(self.size))
        if self.type == tarfile.XGLTYPE:
            pax_headers = tarstream.pax_headers
        else:
            pax_headers = tarstream.pax_headers.copy()
        match = re.search(br"\d+ hdrcharset=([^\n]+)\n", buf)
        if match is not None:
            pax_headers["hdrcharset"] = match.group(1).decode("utf-8")
        hdrcharset = pax_headers.get("hdrcharset")
        if hdrcharset == "BINARY":
            encoding = tarstream.encoding
        else:
            encoding = "utf-8"
        regex = re.compile(br"(\d+) ([^=]+)=")
        pos = 0
        while True:
            match = regex.match(buf, pos)
            if not match:
                break
            length, keyword = match.groups()
            length = int(length)
            if length == 0:
                raise tarfile.InvalidHeaderError("invalid header")
            value = buf[match.end(2) + 1:match.start(1) + length - 1]
            keyword = self._decode_pax_field(keyword, "utf-8", "utf-8", tarstream.errors)
            if keyword in tarfile.PAX_NAME_FIELDS:
                value = self._decode_pax_field(value, encoding, tarstream.encoding, tarstream.errors)
            else:
                value = self._decode_pax_field(value, "utf-8", "utf-8", tarstream.errors)
            pax_headers[keyword] = value
            pos += length
        try:
            next = await self.fromtarfile(tarstream)
        except tarfile.HeaderError as e:
            raise tarfile.SubsequentHeaderError(str(e)) from None
        if "GNU.sparse.map" in pax_headers:
            self._proc_gnusparse_01(next, pax_headers)
        elif "GNU.sparse.size" in pax_headers:
            self._proc_gnusparse_00(next, pax_headers, buf)
        elif pax_headers.get("GNU.sparse.major") == "1" and pax_headers.get("GNU.sparse.minor") == "0":
            await self._proc_gnusparse_10(next, tarstream)
        if self.type in (tarfile.XHDTYPE, tarfile.SOLARIS_XHDTYPE):
            next._apply_pax_info(pax_headers, tarstream.encoding, tarstream.errors)
            next.offset = self.offset
            if "size" in pax_headers:
                offset = next.offset_data
                if next.isreg() or next.type not in tarfile.SUPPORTED_TYPES:
                    offset += next._block(next.size)
                tarstream.offset = offset
        return next

    async def _proc_sparse(self, tarstream):
        structs, isextended, origsize = self._sparse_structs
        del self._sparse_structs
        while isextended:
            buf = await tarstream.stream.read(tarfile.BLOCKSIZE)
            pos = 0
            for i in range(21):
                try:
                    offset = tarfile.nti(buf[pos:pos + 12])
                    numbytes = tarfile.nti(buf[pos + 12:pos + 24])
                except ValueError:
                    break
                if offset and numbytes:
                    structs.append((offset, numbytes))
                pos += 24
            isextended = bool(buf[504])
        self.sparse = structs
        self.offset_data = tarstream.offset
        tarstream.offset = self.offset_data + self._block(self.size)
        self.size = origsize
        return self


class AioTarStream(object):
    debug = 0
    dereference = False
    ignore_zeros = False
    errorlevel = 1
    format = tarfile.DEFAULT_FORMAT
    encoding = tarfile.ENCODING
    errors = None
    tarinfo = AioTarInfo
    fileobject = FileStreamReaderWrapper

    def __init__(self,
                 stream,
                 mode="r",
                 format=None,
                 tarinfo=None,
                 dereference=None,
                 ignore_zeros=None,
                 encoding=None,
                 errors="surrogateescape",
                 pax_headers=None,
                 debug=None,
                 errorlevel=None,
                 copybufsize=None):
        modes = {"r": "rb", "a": "r+b", "w": "wb", "x": "xb"}
        if mode not in modes:
            raise ValueError("mode must be 'r', 'a', 'w' or 'x'")
        self.mode = mode
        self._mode = stream.mode if hasattr(stream, "mode") else modes[mode]
        self.stream = stream
        self.offset = 0
        self.index = 0
        if format is not None:
            self.format = format
        if tarinfo is not None:
            self.tarinfo = tarinfo
        if dereference is not None:
            self.dereference = dereference
        if ignore_zeros is not None:
            self.ignore_zeros = ignore_zeros
        if encoding is not None:
            self.encoding = encoding
        self.errors = errors
        self.pax_headers = pax_headers if pax_headers is not None and self.format == tarfile.PAX_FORMAT else {}
        if debug is not None:
            self.debug = debug
        if errorlevel is not None:
            self.errorlevel = errorlevel
        self.copybufsize = copybufsize
        self.closed = False
        self.members = []
        self._loaded = False
        self.inodes = {}

    async def __aenter__(self):
        self._check()
        try:
            if self.mode == "r":
                self.firstmember = None
                self.firstmember = await self.next()
            if self.mode in ("a", "w", "x"):
                self._loaded = True
                if self.pax_headers:
                    buf = self.tarinfo.create_pax_global_header(self.pax_headers.copy())
                    await self.stream.write(buf)
                    self.offset += len(buf)
            return self
        except BaseException:
            self.closed = True
            raise

    async def __aexit__(self, exc_type, value, traceback):
        if exc_type is None:
            await self.close()
        else:
            self.closed = True

    def __aiter__(self):
        return self

    async def __anext__(self):
        index = self.index
        self.index += 1
        if self._loaded:
            if index < len(self.members):
                return self.members[index]
        else:
            if index == 0 and self.firstmember is not None:
                return await self.next()
            elif index < len(self.members):
                return self.members[index]
            else:
                tarinfo = await self.next()
                if not tarinfo:
                    self._loaded = True
                    raise StopAsyncIteration
                else:
                    return tarinfo

    @classmethod
    def open(cls, stream, mode="r", **kwargs):
        filemode, comptype = mode.split(":", 1) if ':' in mode else (mode, None)
        filemode = mode or "r"
        comptype = comptype or "tar"
        if filemode not in ("r", "w"):
            raise ValueError("mode must be 'r' or 'w'")
        if comptype == "gz":
            stream = GZipStreamWrapper(stream, filemode)
        elif comptype == "bz2":
            stream = BZ2StreamWrapper(stream, filemode)
        elif comptype == "xz":
            stream = LZMAStreamWrapper(stream, filemode)
        elif comptype != "tar":
            raise tarfile.CompressionError("unknown compression type %r" % comptype)
        else:
            stream = BaseStreamWrapper(stream)
        t = cls(stream, filemode, **kwargs)
        return t

    @classmethod
    def taropen(cls, stream, mode="r", **kwargs):
        if mode not in ("r", "a", "w", "x"):
            raise ValueError("mode must be 'r', 'a', 'w' or 'x'")
        return cls(BaseStreamWrapper(stream), mode, **kwargs)

    @classmethod
    def gzopen(cls, stream, mode="r", compresslevel=9, **kwargs):
        if mode not in ("r", "w", "x"):
            raise ValueError("mode must be 'r', 'w' or 'x'")
        return cls(GZipStreamWrapper(stream, mode, compresslevel), mode, **kwargs)

    @classmethod
    def bz2open(cls, stream, mode="r", compresslevel=9, **kwargs):
        if mode not in ("r", "w", "x"):
            raise ValueError("mode must be 'r', 'w' or 'x'")
        return cls(BZ2StreamWrapper(stream, mode, compresslevel), mode, **kwargs)

    @classmethod
    def xzopen(cls, stream, mode="r", preset=None, **kwargs):
        if mode not in ("r", "w", "x"):
            raise ValueError("mode must be 'r', 'w' or 'x'")
        return cls(LZMAStreamWrapper(stream, mode, preset), mode, **kwargs)

    def _check(self, mode=None):
        if self.closed:
            raise OSError("%s is closed" % self.__class__.__name__)
        if mode is not None and self.mode not in mode:
            raise OSError("bad operation for mode %r" % self.mode)

    def _dbg(self, level, msg):
        if level <= self.debug:
            print(msg, file=sys.stderr)

    async def _extract_member(self, tarinfo, targetpath, set_attrs=True, numeric_owner=False):
        targetpath = targetpath.rstrip("/")
        targetpath = targetpath.replace("/", os.sep)
        upperdirs = os.path.dirname(targetpath)
        if upperdirs and not os.path.exists(upperdirs):
            os.makedirs(upperdirs)
        if tarinfo.islnk() or tarinfo.issym():
            self._dbg(1, "%s -> %s" % (tarinfo.name, tarinfo.linkname))
        else:
            self._dbg(1, tarinfo.name)
        if tarinfo.isreg():
            await self.makefile(tarinfo, targetpath)
        elif tarinfo.isdir():
            self.makedir(tarinfo, targetpath)
        elif tarinfo.isfifo():
            self.makefifo(tarinfo, targetpath)
        elif tarinfo.ischr() or tarinfo.isblk():
            self.makedev(tarinfo, targetpath)
        elif tarinfo.islnk() or tarinfo.issym():
            await self.makelink(tarinfo, targetpath)
        elif tarinfo.type not in tarfile.SUPPORTED_TYPES:
            await self.makeunknown(tarinfo, targetpath)
        else:
            await self.makefile(tarinfo, targetpath)
        if set_attrs:
            self.chown(tarinfo, targetpath, numeric_owner)
            if not tarinfo.issym():
                self.chmod(tarinfo, targetpath)
                self.utime(tarinfo, targetpath)

    async def _find_link_target(self, tarinfo):
        if tarinfo.issym():
            linkname = "/".join(filter(None, (os.path.dirname(tarinfo.name), tarinfo.linkname)))
            limit = None
        else:
            linkname = tarinfo.linkname
            limit = tarinfo
        member = await self._getmember(linkname, tarinfo=limit, normalize=True)
        if member is None:
            raise KeyError("linkname %r not found" % linkname)
        return member

    async def _getmember(self, name, tarinfo=None, normalize=False):
        members = await self.getmembers()
        if tarinfo is not None:
            members = members[:members.index(tarinfo)]
        if normalize:
            name = os.path.normpath(name)
        for member in reversed(members):
            if normalize:
                member_name = os.path.normpath(member.name)
            else:
                member_name = member.name
            if name == member_name:
                return member

    async def _load(self):
        while True:
            tarinfo = await self.next()
            if tarinfo is None:
                break
        self._loaded = True

    async def add(self, name, arcname=None, recursive=True, *, filter=None):
        self._check("awx")
        if arcname is None:
            arcname = name
        self._dbg(1, name)
        tarinfo = self.gettarinfo(name, arcname)
        if tarinfo is None:
            self._dbg(1, "tarfile: Unsupported type %r" % name)
            return
        if filter is not None:
            tarinfo = filter(tarinfo)
            if tarinfo is None:
                self._dbg(2, "tarfile: Excluded %r" % name)
                return
        if tarinfo.isreg():
            with bltn_open(name, "rb") as f:
                await self.addfile(tarinfo, f)
        elif tarinfo.isdir():
            await self.addfile(tarinfo)
            if recursive:
                for f in sorted(os.listdir(name)):
                    await self.add(os.path.join(name, f), os.path.join(arcname, f), recursive, filter=filter)
        else:
            await self.addfile(tarinfo)

    async def addfile(self, tarinfo, fileobj=None):
        self._check("awx")
        tarinfo = copy.copy(tarinfo)
        buf = tarinfo.tobuf(self.format, self.encoding, self.errors)
        await self.stream.write(buf)
        self.offset += len(buf)
        bufsize = self.copybufsize
        if fileobj is not None:
            await copyfileobj(fileobj, self.stream, tarinfo.size, bufsize)
            blocks, remainder = divmod(tarinfo.size, tarfile.BLOCKSIZE)
            if remainder > 0:
                await self.stream.write(tarfile.NUL * (tarfile.BLOCKSIZE - remainder))
                blocks += 1
            self.offset += blocks * tarfile.BLOCKSIZE
        self.members.append(tarinfo)

    def chown(self, tarinfo, targetpath, numeric_owner):
        if hasattr(os, "geteuid") and os.geteuid() == 0:
            g = tarinfo.gid
            u = tarinfo.uid
            if not numeric_owner:
                try:
                    if grp:
                        g = grp.getgrnam(tarinfo.gname)[2]
                except KeyError:
                    pass
                try:
                    if pwd:
                        u = pwd.getpwnam(tarinfo.uname)[2]
                except KeyError:
                    pass
            try:
                if tarinfo.issym() and hasattr(os, "lchown"):
                    os.lchown(targetpath, u, g)
                else:
                    os.chown(targetpath, u, g)
            except OSError as e:
                raise tarfile.ExtractError("could not change owner") from e

    def chmod(self, tarinfo, targetpath):
        try:
            os.chmod(targetpath, tarinfo.mode)
        except OSError as e:
            raise tarfile.ExtractError("could not change mode") from e

    async def close(self):
        if self.closed:
            return
        self.closed = True
        try:
            if self.mode in ("a", "w", "x"):
                await self.stream.write(tarfile.NUL * (tarfile.BLOCKSIZE * 2))
                self.offset += (tarfile.BLOCKSIZE * 2)
                blocks, remainder = divmod(self.offset, tarfile.RECORDSIZE)
                if remainder > 0:
                    await self.stream.write(tarfile.NUL * (tarfile.RECORDSIZE - remainder))
        finally:
            await self.stream.close()

    async def extract(self, member, path="", set_attrs=True, *, numeric_owner=False):
        self._check("r")
        tarinfo = await self.getmember(member) if isinstance(member, str) else member
        if tarinfo.islnk():
            tarinfo._link_target = os.path.join(path, tarinfo.linkname)
        try:
            await self._extract_member(tarinfo,
                                       os.path.join(path, tarinfo.name),
                                       set_attrs=set_attrs,
                                       numeric_owner=numeric_owner)
        except OSError as e:
            if self.errorlevel > 0:
                raise
            else:
                if e.filename is None:
                    self._dbg(1, "tarfile: %s" % e.strerror)
                else:
                    self._dbg(1, "tarfile: %s %r" % (e.strerror, e.filename))
        except tarfile.ExtractError as e:
            if self.errorlevel > 1:
                raise
            else:
                self._dbg(1, "tarfile: %s" % e)

    async def extractall(self, path=".", members=None, *, numeric_owner=False):
        directories = []
        if members is None:
            members = self
        for tarinfo in members:
            if tarinfo.isdir():
                directories.append(tarinfo)
                tarinfo = copy.copy(tarinfo)
                tarinfo.mode = 0o700
            await self.extract(tarinfo, path, set_attrs=not tarinfo.isdir(), numeric_owner=numeric_owner)
        directories.sort(key=lambda a: a.name)
        directories.reverse()
        for tarinfo in directories:
            dirpath = os.path.join(path, tarinfo.name)
            try:
                self.chown(tarinfo, dirpath, numeric_owner=numeric_owner)
                self.utime(tarinfo, dirpath)
                self.chmod(tarinfo, dirpath)
            except tarfile.ExtractError as e:
                if self.errorlevel > 1:
                    raise
                else:
                    self._dbg(1, "tarfile: %s" % e)

    async def extractfile(self, member):
        self._check("r")
        tarinfo = await self.getmember(member) if isinstance(member, str) else member
        if tarinfo.isreg() or tarinfo.type not in tarinfo.SUPPORTED_TYPES:
            return self.fileobject(
                stream=self.stream,
                size=tarinfo.size,
                blockinfo=tarinfo.sparse)
        elif tarinfo.islnk() or tarinfo.issym():
            raise tarinfo.StreamError("cannot extract (sym)link as file object")
        else:
            return None

    async def getmember(self, name):
        tarinfo = await self._getmember(name.rstrip('/'))
        if tarinfo is None:
            raise KeyError("filename %r not found" % name)
        return tarinfo

    async def getmembers(self):
        self._check()
        if not self._loaded:
            await self._load()
        return self.members

    async def getnames(self):
        return [tarinfo.name for tarinfo in await self.getmembers()]

    def gettarinfo(self, name=None, arcname=None, fileobj=None):
        self._check("awx")
        if fileobj is not None:
            name = fileobj.name
        if arcname is None:
            arcname = name
        drv, arcname = os.path.splitdrive(arcname)
        arcname = arcname.replace(os.sep, "/")
        arcname = arcname.lstrip("/")
        tarinfo = self.tarinfo()
        if fileobj is None:
            statres = os.lstat(name) if not self.dereference else os.stat(name)
        else:
            statres = os.fstat(fileobj.fileno())
        linkname = ""
        stmd = statres.st_mode
        if stat.S_ISREG(stmd):
            inode = (statres.st_ino, statres.st_dev)
            if not self.dereference and statres.st_nlink > 1 and \
                    inode in self.inodes and arcname != self.inodes[inode]:
                type = tarfile.LNKTYPE
                linkname = self.inodes[inode]
            else:
                type = tarfile.REGTYPE
                if inode[0]:
                    self.inodes[inode] = arcname
        elif stat.S_ISDIR(stmd):
            type = tarfile.DIRTYPE
        elif stat.S_ISFIFO(stmd):
            type = tarfile.FIFOTYPE
        elif stat.S_ISLNK(stmd):
            type = tarfile.SYMTYPE
            linkname = os.readlink(name)
        elif stat.S_ISCHR(stmd):
            type = tarfile.CHRTYPE
        elif stat.S_ISBLK(stmd):
            type = tarfile.BLKTYPE
        else:
            return None
        tarinfo.name = arcname
        tarinfo.mode = stmd
        tarinfo.uid = statres.st_uid
        tarinfo.gid = statres.st_gid
        tarinfo.size = statres.st_size if type == tarfile.REGTYPE else 0
        tarinfo.mtime = statres.st_mtime
        tarinfo.type = type
        tarinfo.linkname = linkname
        if pwd:
            try:
                tarinfo.uname = pwd.getpwuid(tarinfo.uid)[0]
            except KeyError:
                pass
        if grp:
            try:
                tarinfo.gname = grp.getgrgid(tarinfo.gid)[0]
            except KeyError:
                pass

        if type in (tarfile.CHRTYPE, tarfile.BLKTYPE):
            if hasattr(os, "major") and hasattr(os, "minor"):
                tarinfo.devmajor = os.major(statres.st_rdev)
                tarinfo.devminor = os.minor(statres.st_rdev)
        return tarinfo

    def list(self, verbose=True, *, members=None):
        self._check()
        if members is None:
            members = self
        for tarinfo in members:
            if verbose:
                tarfile._safe_print(stat.filemode(tarinfo.mode))
                tarfile._safe_print("%s/%s" % (tarinfo.uname or tarinfo.uid, tarinfo.gname or tarinfo.gid))
                if tarinfo.ischr() or tarinfo.isblk():
                    tarfile._safe_print("%10s" % ("%d,%d" % (tarinfo.devmajor, tarinfo.devminor)))
                else:
                    tarfile._safe_print("%10d" % tarinfo.size)
                tarfile._safe_print("%d-%02d-%02d %02d:%02d:%02d" % time.localtime(tarinfo.mtime)[:6])
            tarfile._safe_print(tarinfo.name + ("/" if tarinfo.isdir() else ""))
            if verbose:
                if tarinfo.issym():
                    tarfile._safe_print("-> " + tarinfo.linkname)
                if tarinfo.islnk():
                    tarfile._safe_print("link to " + tarinfo.linkname)
            print()

    def makedev(self, tarinfo, targetpath):
        if not hasattr(os, "mknod") or not hasattr(os, "makedev"):
            raise tarfile.ExtractError("special devices not supported by system")
        mode = tarinfo.mode
        if tarinfo.isblk():
            mode |= stat.S_IFBLK
        else:
            mode |= stat.S_IFCHR
        os.mknod(targetpath, mode, os.makedev(tarinfo.devmajor, tarinfo.devminor))

    def makedir(self, tarinfo, targetpath):
        try:
            os.mkdir(targetpath, 0o700)
        except FileExistsError:
            pass

    def makefifo(self, tarinfo, targetpath):
        if hasattr(os, "mkfifo"):
            os.mkfifo(targetpath)
        else:
            raise tarfile.ExtractError("fifo not supported by system")

    async def makefile(self, tarinfo, targetpath):
        bufsize = self.copybufsize
        with bltn_open(targetpath, "wb") as target:
            if tarinfo.sparse is not None:
                for offset, size in tarinfo.sparse:
                    target.seek(offset)
                    await copyfileobj(self.stream, target, size, bufsize)
                target.seek(tarinfo.size)
                target.truncate()
            else:
                await copyfileobj(self.stream, target, tarinfo.size, bufsize)

    async def makelink(self, tarinfo, targetpath):
        try:
            if tarinfo.issym():
                if os.path.lexists(targetpath):
                    os.unlink(targetpath)
                os.symlink(tarinfo.linkname, targetpath)
            else:
                if os.path.exists(tarinfo._link_target):
                    os.link(tarinfo._link_target, targetpath)
                else:
                    await self._extract_member(await self._find_link_target(tarinfo), targetpath)
        except tarfile.symlink_exception:
            try:
                await self._extract_member(await self._find_link_target(tarinfo), targetpath)
            except KeyError:
                raise tarfile.ExtractError("unable to resolve link inside archive") from None

    async def makeunknown(self, tarinfo, targetpath):
        await self.makefile(tarinfo, targetpath)
        self._dbg(1, "tarfile: Unknown file type %r, extracted as regular file." % tarinfo.type)

    async def next(self):
        self._check("ra")
        if self.firstmember is not None:
            m = self.firstmember
            self.firstmember = None
            return m
        tarinfo = None
        while True:
            try:
                tarinfo = await self.tarinfo.fromtarfile(self)
            except tarfile.EOFHeaderError as e:
                if self.ignore_zeros:
                    self._dbg(2, "0x%X: %s" % (self.offset, e))
                    self.offset += tarfile.BLOCKSIZE
                    continue
            except tarfile.InvalidHeaderError as e:
                if self.ignore_zeros:
                    self._dbg(2, "0x%X: %s" % (self.offset, e))
                    self.offset += tarfile.BLOCKSIZE
                    continue
                elif self.offset == 0:
                    raise tarfile.ReadError(str(e)) from None
            except tarfile.EmptyHeaderError:
                if self.offset == 0:
                    raise tarfile.ReadError("empty file") from None
            except tarfile.TruncatedHeaderError as e:
                if self.offset == 0:
                    raise tarfile.ReadError(str(e)) from None
            except tarfile.SubsequentHeaderError as e:
                raise tarfile.ReadError(str(e)) from None
            except Exception as e:
                try:
                    import zlib
                    if isinstance(e, zlib.error):
                        raise tarfile.ReadError(f'zlib error: {e}') from None
                    else:
                        raise e
                except ImportError:
                    raise e
            break
        if tarinfo is not None:
            self.members.append(tarinfo)
        else:
            self._loaded = True
        return tarinfo

    def utime(self, tarinfo, targetpath):
        if not hasattr(os, 'utime'):
            return
        try:
            os.utime(targetpath, (tarinfo.mtime, tarinfo.mtime))
        except OSError as e:
            raise tarfile.ExtractError("could not change modification time") from e


open = AioTarStream.open



from __future__ import annotations

import asyncio
import os
from typing import TYPE_CHECKING

import pkg_resources

from streamflow.core.context import StreamFlowContext
from streamflow.core.data import DataType
from streamflow.core.scheduling import Hardware, JobAllocation, Policy
from streamflow.core.workflow import Status
from streamflow.workflow.token import FileToken

if TYPE_CHECKING:
    from streamflow.core.scheduling import Location, LocationAllocation
    from streamflow.core.workflow import Job
    from typing import MutableMapping, Optional


def _is_valid(current_location: str,
              hardware_requirement: Hardware,
              available_locations: MutableMapping[str, Location],
              jobs: MutableMapping[str, JobAllocation],
              locations: MutableMapping[str, LocationAllocation]) -> bool:
    location_obj = available_locations[current_location]
    running_jobs = list(
        filter(lambda x: jobs[x].status == Status.RUNNING or jobs[x].status == Status.FIREABLE,
               locations[current_location].jobs)) if current_location in locations else []
    # If location is segmentable and job provides requirements, compute the used amount of locations
    if location_obj.hardware is not None and hardware_requirement is not None:
        used_hardware = sum((jobs[j].hardware for j in running_jobs), start=hardware_requirement.__class__())
        if (location_obj.hardware - used_hardware) >= hardware_requirement:
            return True
        else:
            return False
    # If location is segmentable but job does not provide requirements, treat it as null-weighted
    elif location_obj.hardware is not None:
        return True
    # Otherwise, simply compute the number of allocated slots
    else:
        return len(running_jobs) < available_locations[current_location].slots


class DataLocalityPolicy(Policy):

    async def get_location(self,
                           context: StreamFlowContext,
                           job: Job,
                           deployment: str,
                           hardware_requirement: Hardware,
                           available_locations: MutableMapping[str, Location],
                           jobs: MutableMapping[str, JobAllocation],
                           locations: MutableMapping[str, LocationAllocation]) -> Optional[str]:
        valid_locations = list(available_locations.keys())
        # For each input token sorted by weight
        weights = {k: v for k, v in zip(job.inputs, await asyncio.gather(*(
            asyncio.create_task(t.get_weight(context)) for t in job.inputs.values())))}
        for name, token in sorted(job.inputs.items(), key=lambda item: weights[item[0]], reverse=True):
            related_locations = set()
            # For FileTokens, retrieve related locations
            if isinstance(token, FileToken):
                for path in await token.get_paths(context):
                    related_locations.update([loc.location for loc in context.data_manager.get_data_locations(
                        path, deployment, DataType.PRIMARY)])
            # Check if one of the related locations is free
            for current_location in related_locations:
                if current_location in valid_locations:
                    if _is_valid(
                            current_location=current_location,
                            hardware_requirement=hardware_requirement,
                            available_locations=available_locations,
                            jobs=jobs,
                            locations=locations):
                        return current_location
                    else:
                        valid_locations.remove(current_location)
        # If a data-related allocation is not possible, assign a location among the remaining free ones
        for location in valid_locations:
            if _is_valid(
                    current_location=location,
                    hardware_requirement=hardware_requirement,
                    available_locations=available_locations,
                    jobs=jobs,
                    locations=locations):
                return location
        # If there are no available locations, return None
        return None

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'data_locality.json'))



from __future__ import annotations

from collections import deque
from typing import Any, AsyncIterable, MutableMapping, MutableSequence, Union, cast

from streamflow.core import utils
from streamflow.core.context import StreamFlowContext
from streamflow.core.exception import WorkflowExecutionException
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.workflow import Token, Workflow
from streamflow.workflow.step import Combinator
from streamflow.workflow.token import IterationTerminationToken


def _add_to_list(token: Union[Token, MutableMapping[str, Token]],
                 token_values: MutableMapping[str, MutableMapping[str, MutableSequence[Any]]],
                 port_name: str,
                 depth: int = 0):
    tag = utils.get_tag(token.values()) if isinstance(token, MutableMapping) else token.tag
    if depth:
        tag = '.'.join(tag.split('.')[:-depth])
    for key in list(token_values.keys()):
        if tag == key:
            continue
        elif key.startswith(tag):
            _add_to_port(token, token_values[key], port_name)
        elif tag.startswith(key):
            if tag not in token_values:
                token_values[tag] = {}
            for p in token_values[key]:
                for t in token_values[key][p]:
                    _add_to_port(t, token_values[tag], p)
    if tag not in token_values:
        token_values[tag] = {}
    _add_to_port(token, token_values[tag], port_name)


def _add_to_port(token: Union[Token, MutableMapping[str, Token]],
                 tag_values: MutableMapping[str, MutableSequence[Any]],
                 port_name: str):
    if port_name not in tag_values:
        tag_values[port_name] = deque()
    tag_values[port_name].append(token)


class CartesianProductCombinator(Combinator):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 depth: int = 1):
        super().__init__(name, workflow)
        self.depth: int = depth
        self.token_values: MutableMapping[str, MutableMapping[str, MutableSequence[Any]]] = {}

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CartesianProductCombinator:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            depth=row['depth'])

    async def _product(self,
                       port_name: str,
                       token: Union[Token, MutableSequence[Token]]) -> AsyncIterable[MutableMapping[str, Token]]:
        # Get all combinations of the new element with the others
        tag = '.'.join(token.tag.split('.')[:-self.depth])
        if len(self.token_values[tag]) == len(self.items):
            cartesian_product = utils.dict_product(
                **{k: [token] if k == port_name else v for k, v in self.token_values[tag].items()})
            # Return all combination schemas
            for config in cartesian_product:
                schema = {}
                for key in self.items:
                    if key in self.combinators:
                        schema = {**schema, **config[key]}
                    else:
                        schema[key] = config[key]
                        schema[key] = config[key]
                suffix = [t.tag.split('.')[-1] for t in schema.values()]
                schema = {k: t.retag('.'.join(t.tag.split('.')[:-1] + suffix)) for k, t in schema.items()}
                yield schema

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'depth': self.depth}}

    async def combine(self,
                      port_name: str,
                      token: Token) -> AsyncIterable[MutableMapping[str, Token]]:
        # If port is associated to an inner combinator, call it and put shcemas in their related list
        if c := self.get_combinator(port_name):
            async for schema in cast(AsyncIterable, c.combine(port_name, token)):
                _add_to_list(schema, self.token_values, c.name, self.depth)
                async for product in self._product(port_name, token):
                    yield product
        # If port is associated directly with the current combinator, put the token in the list
        elif port_name in self.items:
            _add_to_list(token, self.token_values, port_name, self.depth)
            async for product in self._product(port_name, token):
                yield product
        # Otherwise throw Exception
        else:
            raise WorkflowExecutionException("No item to combine for token '{}'.".format(port_name))


class DotProductCombinator(Combinator):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__(name, workflow)
        self.token_values: MutableMapping[str, MutableMapping[str, MutableSequence[Any]]] = {}

    async def _product(self) -> AsyncIterable[MutableMapping[str, Token]]:
        # Check if some complete input sets are available
        for tag in list(self.token_values):
            if len(self.token_values[tag]) == len(self.items):
                num_items = min(len(i) for i in self.token_values[tag].values())
                for i in range(num_items):
                    # Return the relative combination schema
                    schema = {}
                    for key, elements in self.token_values[tag].items():
                        element = elements.pop()
                        if key in self.combinators:
                            schema = {**schema, **element}
                        else:
                            schema[key] = element
                    tag = utils.get_tag(schema.values())
                    schema = {k: t.retag(tag) for k, t in schema.items()}
                    yield schema

    async def combine(self,
                      port_name: str,
                      token: Token) -> AsyncIterable[MutableMapping[str, Token]]:
        # If port is associated to an inner combinator, call it and put shcemas in their related list
        if c := self.get_combinator(port_name):
            async for schema in cast(AsyncIterable, c.combine(port_name, token)):
                _add_to_list(schema, self.token_values, c.name)
                async for product in self._product():
                    yield product
        # If port is associated directly with the current combinator, put the token in the list
        elif port_name in self.items:
            _add_to_list(token, self.token_values, port_name)
            async for product in self._product():
                yield product
        # Otherwise throw Exception
        else:
            raise WorkflowExecutionException("No item to combine for token '{}'.".format(port_name))


class LoopCombinator(DotProductCombinator):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__(name, workflow)
        self.iteration_map: MutableMapping[str, int] = {}

    async def _product(self) -> AsyncIterable[MutableMapping[str, Token]]:
        async for schema in super()._product():
            tag = utils.get_tag(schema.values())
            prefix = '.'.join(tag.split('.')[:-1])
            if prefix not in self.iteration_map:
                self.iteration_map[tag] = 0
                tag = '.'.join(tag.split('.') + ['0'])
            else:
                self.iteration_map[prefix] += 1
                tag = '.'.join(tag.split('.')[:-1] + [str(self.iteration_map[prefix])])
            schema = {k: t.retag(tag) for k, t in schema.items()}
            yield schema


class LoopTerminationCombinator(DotProductCombinator):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__(name, workflow)
        self.output_items: MutableSequence[str] = []
        self.token_values: MutableMapping[str, MutableMapping[str, MutableSequence[Any]]] = {}

    def add_output_item(self, item: str) -> None:
        self.output_items.append(item)

    async def _product(self) -> AsyncIterable[MutableMapping[str, Token]]:
        async for schema in super()._product():
            tag = utils.get_tag(schema.values())
            yield {k: IterationTerminationToken(tag=tag) for k in self.output_items}



from typing import MutableMapping

from streamflow.core.context import StreamFlowContext
from streamflow.core.deployment import DeploymentConfig, Target
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.workflow import Port, Step, Token, Workflow


class DefaultDatabaseLoadingContext(DatabaseLoadingContext):

    def __init__(self):
        super().__init__()
        self._deployment_configs: MutableMapping[int, DeploymentConfig] = {}
        self._ports: MutableMapping[int, Port] = {}
        self._steps: MutableMapping[int, Step] = {}
        self._targets: MutableMapping[int, Target] = {}
        self._tokens: MutableMapping[int, Token] = {}
        self._workflows: MutableMapping[int, Workflow] = {}

    def add_deployment(self,
                       persistent_id: int,
                       deployment: DeploymentConfig):
        self._deployment_configs[persistent_id] = deployment

    def add_port(self,
                 persistent_id: int,
                 port: Port):
        self._ports[persistent_id] = port

    def add_step(self,
                 persistent_id: int,
                 step: Step):
        self._steps[persistent_id] = step

    def add_target(self,
                   persistent_id: int,
                   target: Target):
        self._targets[persistent_id] = target

    def add_token(self,
                  persistent_id: int,
                  token: Token):
        self._tokens[persistent_id] = token

    def add_workflow(self,
                     persistent_id: int,
                     workflow: Workflow):
        self._workflows[persistent_id] = workflow

    async def load_deployment(self,
                              context: StreamFlowContext,
                              persistent_id: int):
        return self._deployment_configs.get(persistent_id) or await DeploymentConfig.load(context, persistent_id, self)

    async def load_port(self,
                        context: StreamFlowContext,
                        persistent_id: int):
        return self._ports.get(persistent_id) or await Port.load(context, persistent_id, self)

    async def load_step(self,
                        context: StreamFlowContext,
                        persistent_id: int):
        return self._steps.get(persistent_id) or await Step.load(context, persistent_id, self)

    async def load_target(self,
                          context: StreamFlowContext,
                          persistent_id: int):
        return self._targets.get(persistent_id) or await Target.load(context, persistent_id, self)

    async def load_token(self,
                         context: StreamFlowContext,
                         persistent_id: int):
        return self._tokens.get(persistent_id) or await Token.load(context, persistent_id, self)

    async def load_workflow(self,
                            context: StreamFlowContext,
                            persistent_id: int):
        return self._workflows.get(persistent_id) or await Workflow.load(context, persistent_id, self)



from __future__ import annotations

import asyncio
from typing import Any, Callable, MutableMapping, MutableSequence, Optional, Type, cast

import cwltool.builder
from rdflib import Graph
from schema_salad.exceptions import ValidationException

from streamflow.core.context import StreamFlowContext
from streamflow.core.data import LOCAL_LOCATION
from streamflow.core.deployment import Connector, Target
from streamflow.core.exception import WorkflowDefinitionException, WorkflowExecutionException
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.utils import flatten_list, get_path_processor, get_tag
from streamflow.core.workflow import CommandOutput, CommandOutputProcessor, Job, Status, Token, TokenProcessor, Workflow
from streamflow.cwl import utils
from streamflow.cwl.command import CWLCommandOutput
from streamflow.cwl.token import CWLFileToken
from streamflow.cwl.utils import LoadListing, SecondaryFile
from streamflow.workflow.token import ListToken, ObjectToken


def _check_default_processor(processor: CWLCommandOutputProcessor, token_value: Any):
    try:
        utils.check_token_type(
            name=processor.name,
            token_value=token_value,
            token_type=processor.token_type,
            enum_symbols=processor.enum_symbols,
            optional=processor.optional)
        return True
    except WorkflowExecutionException:
        return False


class CWLTokenProcessor(TokenProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 token_type: Optional[str] = None,
                 check_type: bool = True,
                 enum_symbols: Optional[MutableSequence[str]] = None,
                 expression_lib: Optional[MutableSequence[str]] = None,
                 file_format: Optional[str] = None,
                 format_graph: Optional[Graph] = None,
                 full_js: bool = False,
                 load_contents: Optional[bool] = None,
                 load_listing: Optional[LoadListing] = None,
                 only_propagate_secondary_files: bool = True,
                 optional: bool = False,
                 secondary_files: Optional[MutableSequence[SecondaryFile]] = None,
                 streamable: bool = False):
        super().__init__(name, workflow)
        self.token_type: str = token_type
        self.check_type: bool = check_type
        self.enum_symbols: Optional[MutableSequence[str]] = enum_symbols
        self.expression_lib: Optional[MutableSequence[str]] = expression_lib
        self.file_format: Optional[str] = file_format
        self.format_graph: Optional[Graph] = format_graph
        self.full_js: bool = full_js
        self.load_contents: Optional[bool] = load_contents
        self.load_listing: Optional[LoadListing] = load_listing
        self.only_propagate_secondary_files: bool = only_propagate_secondary_files
        self.optional: bool = optional
        self.secondary_files: MutableSequence[SecondaryFile] = secondary_files or []
        self.streamable: bool = streamable

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CWLTokenProcessor:
        format_graph = Graph()
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            token_type=row['token_type'],
            check_type=row['check_type'],
            enum_symbols=row['enum_symbols'],
            expression_lib=row['expression_lib'],
            file_format=row['file_format'],
            format_graph=(format_graph.parse(data=row['format_graph']) if row['format_graph'] is not None
                          else format_graph),
            full_js=row['full_js'],
            load_contents=row['load_contents'],
            load_listing=LoadListing(row['load_listing']) if row['load_listing'] is not None else None,
            only_propagate_secondary_files=row['only_propagate_secondary_files'],
            optional=row['optional'],
            secondary_files=[SecondaryFile(sf['pattern'], sf['required']) for sf in row['secondary_files']],
            streamable=row['streamable'])

    async def _process_file_token(self, inputs: MutableMapping[str, Token], token_value: Any):
        # Build context
        context = utils.build_context(inputs)
        # Get path
        filepath = utils.get_path_from_token(token_value)
        # Check file format if present
        if self.file_format:
            context = {**context, **{'self': token_value}}
            input_formats = utils.eval_expression(
                expression=self.file_format,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib)
            try:
                cwltool.builder.check_format(token_value, input_formats, self.format_graph)
            except ValidationException as e:
                raise WorkflowExecutionException(e.message) from e
        # If file exists, get coordinates
        if filepath and (
                data_location := self.workflow.context.data_manager.get_source_location(filepath, LOCAL_LOCATION)):
            connector = self.workflow.context.deployment_manager.get_connector(data_location.deployment)
            path_processor = utils.get_path_processor(connector)
            location = data_location.location
            base_path = path_processor.normpath(data_location.path[:-len(data_location.relpath)])
            # Process file contents
            token_value = await utils.update_file_token(
                context=self.workflow.context,
                connector=connector,
                location=location,
                token_value=token_value,
                load_contents=self.load_contents,
                load_listing=self.load_listing)
            # Process secondary files
            if token_value.get('secondaryFiles'):
                initial_paths = [utils.get_path_from_token(sf) for sf in token_value['secondaryFiles']]
                sf_map = dict(zip(initial_paths, await asyncio.gather(*(asyncio.create_task(
                    utils.update_file_token(
                        context=self.workflow.context,
                        connector=connector,
                        location=location,
                        token_value=sf,
                        load_contents=self.load_contents,
                        load_listing=self.load_listing
                    )) for sf in token_value['secondaryFiles']))))
            else:
                sf_map = {}
            if self.secondary_files:
                sf_context = {**context, 'self': token_value}
                await utils.process_secondary_files(
                    context=self.workflow.context,
                    secondary_files=self.secondary_files,
                    sf_map=sf_map,
                    js_context=sf_context,
                    full_js=self.full_js,
                    expression_lib=self.expression_lib,
                    connector=connector,
                    locations=[location],
                    token_value=token_value,
                    load_contents=self.load_contents,
                    load_listing=self.load_listing,
                    only_retrieve_from_token=self.only_propagate_secondary_files)
            # Add all secondary files to the token
            if sf_map:
                token_value['secondaryFiles'] = list(sf_map.values())
            # Register path
            await utils.register_data(
                context=self.workflow.context,
                connector=connector,
                locations=[location],
                token_value=token_value,
                base_path=base_path)
        # Return token value
        return token_value

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'token_type': self.token_type,
                   'check_type': self.check_type,
                   'enum_symbols': self.enum_symbols,
                   'expression_lib': self.expression_lib,
                   'file_format': self.file_format,
                   'format_graph': self.format_graph.serialize() if self.format_graph else None,
                   'full_js': self.full_js,
                   'load_contents': self.load_contents,
                   'load_listing': self.load_listing.value if self.load_listing else None,
                   'only_propagate_secondary_files': self.only_propagate_secondary_files,
                   'optional': self.optional,
                   'secondary_files': await asyncio.gather(*(asyncio.create_task(s.save(context))
                                                             for s in self.secondary_files)),
                   'streamable': self.streamable}}

    async def process(self, inputs: MutableMapping[str, Token], token: Token) -> Token:
        # If value is token, propagate the process call
        if isinstance(token.value, Token):
            return token.update(await self.process(inputs, token.value))
        # Check type
        if self.check_type and self.token_type:
            utils.check_token_type(
                name=self.name,
                token_value=token.value,
                token_type=self.token_type,
                enum_symbols=self.enum_symbols,
                optional=self.optional)
        # Process token
        if utils.get_token_class(token.value) in ['File', 'Directory']:
            return token.update(await self._process_file_token(inputs, token.value))
        else:
            return token


class CWLCommandOutputProcessor(CommandOutputProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 target: Optional[Target] = None,
                 token_type: Optional[str] = None,
                 enum_symbols: Optional[MutableSequence[str]] = None,
                 expression_lib: Optional[MutableSequence[str]] = None,
                 file_format: Optional[str] = None,
                 full_js: bool = False,
                 glob: Optional[str] = None,
                 load_contents: bool = False,
                 load_listing: LoadListing = LoadListing.no_listing,
                 optional: bool = False,
                 output_eval: Optional[str] = None,
                 secondary_files: Optional[MutableSequence[SecondaryFile]] = None,
                 streamable: bool = False):
        super().__init__(name, workflow, target)
        self.token_type: str = token_type
        self.enum_symbols: Optional[MutableSequence[str]] = enum_symbols
        self.expression_lib: Optional[MutableSequence[str]] = expression_lib
        self.file_format: Optional[str] = file_format
        self.full_js: bool = full_js
        self.glob: Optional[str] = glob
        self.load_contents: bool = load_contents
        self.load_listing: LoadListing = load_listing
        self.optional: bool = optional
        self.output_eval: Optional[str] = output_eval
        self.secondary_files: MutableSequence[SecondaryFile] = secondary_files or []
        self.streamable: bool = streamable

    async def _build_token(self,
                           job: Job,
                           connector: Optional[Connector],
                           context: MutableMapping[str, Any],
                           token_value: Any) -> Token:
        if isinstance(token_value, MutableMapping):
            if utils.get_token_class(token_value) in ['File', 'Directory']:
                connector = self._get_connector(connector, job)
                locations = await self._get_locations(connector, job)
                # Register path
                await utils.register_data(
                    context=self.workflow.context,
                    connector=connector,
                    locations=locations,
                    token_value=token_value,
                    base_path=(self.target.workdir if self.target else job.output_directory))
                # Process file format
                if self.file_format:
                    context = {**context, **{'self': token_value}}
                    token_value['format'] = utils.eval_expression(
                        expression=self.file_format,
                        context=context,
                        full_js=self.full_js,
                        expression_lib=self.expression_lib)
                return CWLFileToken(
                    value=token_value,
                    tag=get_tag(job.inputs.values()))
            else:
                token_tasks = {k: asyncio.create_task(
                    self._build_token(job, connector, context, v)) for k, v in token_value.items()}
                return ObjectToken(
                    value=dict(zip(token_tasks.keys(), await asyncio.gather(*token_tasks.values()))),
                    tag=get_tag(job.inputs.values()))
        elif isinstance(token_value, MutableSequence):
            return ListToken(
                value=await asyncio.gather(*(asyncio.create_task(
                    self._build_token(job, connector, context, t)) for t in token_value)),
                tag=get_tag(job.inputs.values()))
        else:
            return Token(
                value=token_value,
                tag=get_tag(job.inputs.values()))

    async def _process_command_output(self,
                                      job: Job,
                                      command_output: CWLCommandOutput,
                                      connector: Optional[Connector],
                                      context: MutableMapping[str, Any],
                                      output_directory: str):
        connector = self._get_connector(connector, job)
        locations = await self._get_locations(connector, job)
        path_processor = get_path_processor(connector)
        token_value = command_output.value
        # If `token_value` is a dictionary, directly extract the token value from it
        if isinstance(token_value, MutableMapping) and self.name in token_value:
            token = token_value[self.name]
            return await utils.build_token_value(
                context=self.workflow.context,
                js_context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib,
                secondary_files=self.secondary_files,
                connector=connector,
                locations=locations,
                token_value=token,
                load_contents=self.load_contents,
                load_listing=self.load_listing)
        # Otherwise, generate the output object as described in `outputs` field
        if self.glob is not None:
            # Adjust glob path
            if '$(' in self.glob or '${' in self.glob:
                globpath = utils.eval_expression(
                    expression=self.glob,
                    context=context,
                    full_js=self.full_js,
                    expression_lib=self.expression_lib)
            else:
                globpath = self.glob
            # Resolve glob
            resolve_tasks = []
            for location in locations:
                globpath = globpath if isinstance(globpath, MutableSequence) else [globpath]
                for path in globpath:
                    resolve_tasks.append(utils.expand_glob(
                        connector=connector,
                        workflow=self.workflow,
                        location=location,
                        input_directory=self.target.workdir if self.target else job.input_directory,
                        output_directory=self.target.workdir if self.target else job.output_directory,
                        tmp_directory=self.target.workdir if self.target else job.tmp_directory,
                        path=(path_processor.join(output_directory, path) if not path_processor.isabs(path)
                              else path)))
            paths, effective_paths = ([list(x) for x in zip(*flatten_list(await asyncio.gather(*resolve_tasks)))]
                                      or [[], []])
            # Get token class from paths
            class_tasks = [asyncio.create_task(utils.get_class_from_path(p, job, self.workflow.context))
                           for p in effective_paths]
            paths = [{'path': p, 'class': c} for p, c in zip(paths, await asyncio.gather(*class_tasks))]
            # If evaluation is not needed, simply return paths as token value
            if self.output_eval is None:
                token_list = await utils.build_token_value(
                    context=self.workflow.context,
                    js_context=context,
                    full_js=self.full_js,
                    expression_lib=self.expression_lib,
                    secondary_files=self.secondary_files,
                    connector=connector,
                    locations=locations,
                    token_value=paths,
                    load_contents=self.load_contents,
                    load_listing=self.load_listing)
                return token_list if len(token_list) > 1 else token_list[0] if len(token_list) == 1 else None
            # Otherwise, fill context['self'] with glob data and proceed
            else:
                context['self'] = await utils.build_token_value(
                    context=self.workflow.context,
                    js_context=context,
                    full_js=self.full_js,
                    expression_lib=self.expression_lib,
                    secondary_files=self.secondary_files,
                    connector=connector,
                    locations=locations,
                    token_value=paths,
                    load_contents=self.load_contents,
                    load_listing=self.load_listing)
        if self.output_eval is not None:
            # Fill context with exit code
            context['runtime']['exitCode'] = command_output.exit_code
            # Evaluate output
            token = utils.eval_expression(
                expression=self.output_eval,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib)
            # Build token
            return await utils.build_token_value(
                context=self.workflow.context,
                js_context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib,
                secondary_files=self.secondary_files,
                connector=connector,
                locations=locations,
                token_value=token,
                load_contents=self.load_contents,
                load_listing=self.load_listing)
        # As the default value (no return path is met in previous code), simply process the command output
        return await utils.build_token_value(
            context=self.workflow.context,
            js_context=context,
            full_js=self.full_js,
            expression_lib=self.expression_lib,
            secondary_files=self.secondary_files,
            connector=connector,
            locations=locations,
            token_value=token_value,
            load_contents=self.load_contents,
            load_listing=self.load_listing)

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'token_type': self.token_type,
                   'enum_symbols': self.enum_symbols,
                   'expression_lib': self.expression_lib,
                   'file_format': self.file_format,
                   'full_js': self.full_js,
                   'glob': self.glob,
                   'load_contents': self.load_contents,
                   'load_listing': self.load_listing.value if self.load_listing else None,
                   'optional': self.optional,
                   'output_eval': self.output_eval,
                   'secondary_files': await asyncio.gather(*(asyncio.create_task(s.save(context))
                                                             for s in self.secondary_files)),
                   'streamable': self.streamable}}

    async def process(self,
                      job: Job,
                      command_output: CWLCommandOutput,
                      connector: Optional[Connector] = None) -> Optional[Token]:
        if command_output.status == Status.SKIPPED:
            return None
        else:
            # Remap output and tmp directories when target is specified
            output_directory = self.target.workdir if self.target else job.output_directory
            tmp_directory = self.target.workdir if self.target else job.tmp_directory
            # Retrieve token value
            context = utils.build_context(
                inputs=job.inputs,
                output_directory=output_directory,
                tmp_directory=tmp_directory,
                hardware=self.workflow.context.scheduler.get_hardware(job.name))
            return await self._build_token(
                job, connector, context, await self._process_command_output(
                    job, command_output, connector, context, output_directory))


class CWLMapTokenProcessor(TokenProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 processor: TokenProcessor):
        super().__init__(name, workflow)
        self.processor: TokenProcessor = processor

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CWLMapTokenProcessor:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            processor=await TokenProcessor.load(context, row['processor'], loading_context))

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'processor': await self.processor.save(context)}}

    async def process(self, inputs: MutableMapping[str, Token], token: Token) -> Token:
        # If value is token, propagate the process call
        if isinstance(token.value, Token):
            return token.update(await self.process(inputs, token.value))
        # Check if token value is a list
        if not isinstance(token, ListToken):
            raise WorkflowDefinitionException(
                "Invalid value {} for token {}: it should be an array".format(token.value, self.name))
        # Propagate evaluation to the inner processor
        return token.update(await asyncio.gather(*(asyncio.create_task(
            self.processor.process(inputs, v)) for v in token.value)))


class CWLMapCommandOutputProcessor(CommandOutputProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 processor: CommandOutputProcessor):
        super().__init__(name, workflow)
        self.processor: CommandOutputProcessor = processor

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CommandOutputProcessor:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            processor=await CommandOutputProcessor.load(context, row['processor'], loading_context))

    async def process(self,
                      job: Job,
                      command_output: CommandOutput,
                      connector: Optional[Connector] = None) -> Optional[Token]:
        if isinstance(command_output.value, MutableMapping) and self.name in command_output.value:
            value = command_output.value[self.name]
            command_output = (command_output.update([{self.name: v} for v in value])
                              if isinstance(value, MutableMapping) else command_output.update(value))
        if isinstance(command_output.value, MutableSequence):
            token = ListToken(
                value=await asyncio.gather(*(asyncio.create_task(
                    self.processor.process(job, command_output.update(value), connector)
                ) for value in command_output.value)),
                tag=get_tag(job.inputs.values()))
        else:
            token = await self.processor.process(job, command_output, connector)
        if not isinstance(token, ListToken):
            token = ListToken(
                value=[token],
                tag=token.tag)
        return token.update(token.value)

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'processor': await self.processor.save(context)}}


class CWLObjectTokenProcessor(TokenProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 processors: MutableMapping[str, TokenProcessor]):
        super().__init__(name, workflow)
        self.processors: MutableMapping[str, TokenProcessor] = processors

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CWLObjectTokenProcessor:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            processors={k: v for k, v in zip(
                row['processors'].keys(),
                await asyncio.gather(*(asyncio.create_task(TokenProcessor.load(context, v, loading_context))
                                       for v in row['processors'].values())))})

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'processors': {k: v for k, v in zip(
                    self.processors.keys(),
                    await asyncio.gather(*(asyncio.create_task(p.save(context))
                                           for p in self.processors.values())))}}}

    async def process(self, inputs: MutableMapping[str, Token], token: Token) -> Token:
        # If value is token, propagate the process call
        if isinstance(token.value, Token):
            return token.update(await self.process(inputs, token.value))
        # Check if token value is a dictionary
        if not isinstance(token.value, MutableMapping):
            raise WorkflowDefinitionException(
                "Invalid value {} for token {}: it should be a record".format(token.value, self.name))
        # Propagate evaluation to the inner processors
        return token.update(dict(zip(token.value, [t.value for t in await asyncio.gather(*(asyncio.create_task(
            self.processors[k].process(inputs, token.update(v))) for k, v in token.value.items()))])))


class CWLObjectCommandOutputProcessor(CommandOutputProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 processors: MutableMapping[str, CommandOutputProcessor],
                 expression_lib: Optional[MutableSequence[str]] = None,
                 full_js: bool = False,
                 output_eval: Optional[str] = None):
        super().__init__(name, workflow)
        self.processors: MutableMapping[str, CommandOutputProcessor] = processors
        self.expression_lib: Optional[MutableSequence[str]] = expression_lib
        self.full_js: bool = full_js
        self.output_eval: Optional[str] = output_eval

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'processors': {k: v for k, v in zip(
                    self.processors.keys(),
                    await asyncio.gather(*(asyncio.create_task(p.save(context))
                                           for p in self.processors.values())))},
                   'expression_lib': self.expression_lib,
                   'full_js': self.full_js,
                   'output_eval': self.output_eval}}

    async def process(self,
                      job: Job,
                      command_output: CommandOutput,
                      connector: Optional[Connector] = None) -> Optional[Token]:
        # Remap output and tmp directories when target is specified
        output_directory = self.target.workdir if self.target else job.output_directory
        tmp_directory = self.target.workdir if self.target else job.tmp_directory
        if self.output_eval:
            # Build context and fill it with exit code
            context = utils.build_context(
                inputs=job.inputs,
                output_directory=output_directory,
                tmp_directory=tmp_directory,
                hardware=self.workflow.context.scheduler.get_hardware(job.name))
            context['runtime']['exitCode'] = cast(CWLCommandOutput, command_output).exit_code
            # Evaluate output
            command_output = command_output.update(utils.eval_expression(
                expression=self.output_eval,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib))
        if isinstance(command_output.value, MutableMapping):
            if self.name in command_output.value:
                return await self.process(
                    job, command_output.update(command_output.value[self.name]), connector)
            else:
                token_tasks = {k: asyncio.create_task(p.process(
                    job, command_output.update(command_output.value[k]), connector))
                    for k, p in self.processors.items() if k in command_output.value}
                return ObjectToken(
                    value=dict(
                        zip(token_tasks.keys(), await asyncio.gather(*token_tasks.values()))),
                    tag=get_tag(job.inputs.values()))
        else:
            token_tasks = {k: asyncio.create_task(p.process(job, command_output, connector))
                           for k, p in self.processors.items()}
            return ObjectToken(
                value=dict(
                    zip(token_tasks.keys(), await asyncio.gather(*token_tasks.values()))),
                tag=get_tag(job.inputs.values()))


class CWLUnionTokenProcessor(TokenProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 processors: MutableSequence[TokenProcessor]):
        super().__init__(name, workflow)
        self.processors: MutableSequence[TokenProcessor] = processors
        self.check_processor: MutableMapping[Type[TokenProcessor], Callable] = {
            CWLTokenProcessor: self._check_default_processor,
            CWLObjectTokenProcessor: self._check_object_processor,
            CWLMapTokenProcessor: self._check_map_processor,
            CWLUnionTokenProcessor: self._check_union_processor}

    # noinspection PyMethodMayBeStatic
    def _check_default_processor(self, processor: CWLCommandOutputProcessor, token_value: Any):
        try:
            utils.check_token_type(
                name=processor.name,
                token_value=token_value,
                token_type=processor.token_type,
                enum_symbols=processor.enum_symbols,
                optional=processor.optional)
            return True
        except WorkflowExecutionException:
            return False

    def _check_map_processor(self, processor: CWLMapTokenProcessor, token_value: Any):
        return isinstance(token_value, MutableSequence) and self.check_processor[type(processor.processor)](
            processor.processor, token_value[0])

    def _check_object_processor(self, processor: CWLObjectTokenProcessor, token_value: Any):
        return (isinstance(token_value, MutableMapping) and
                all(k in processor.processors and
                    self.check_processor[type(processor.processors[k])](processor.processors[k], v)
                    for k, v in token_value.items()))

    def _check_union_processor(self, processor: CWLUnionTokenProcessor, token_value: Any):
        return any(self.check_processor[type(p)](p, token_value) for p in processor.processors)

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CWLUnionTokenProcessor:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            processors=cast(MutableSequence[TokenProcessor], await asyncio.gather(*(asyncio.create_task(
                TokenProcessor.load(context, p, loading_context)) for p in row['processors']))))

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'processors': await asyncio.gather(*(asyncio.create_task(v.save(context))
                                                        for v in self.processors))}}

    def get_processor(self, token_value: Any) -> TokenProcessor:
        for processor in self.processors:
            if self.check_processor[type(processor)](processor, token_value):
                return processor
        raise WorkflowDefinitionException("No suitable command output processors for value " + str(token_value))

    async def process(self, inputs: MutableMapping[str, Token], token: Token) -> Token:
        # If value is token, propagate the process call
        if isinstance(token.value, Token):
            return token.update(await self.process(inputs, token.value))
        # Select the correct processor for the evaluation
        processor = self.get_processor(token.value)
        # Propagate evaluation to the selected processor
        return await processor.process(inputs, token)


class CWLUnionCommandOutputProcessor(CommandOutputProcessor):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 processors: MutableSequence[CommandOutputProcessor]):
        super().__init__(name, workflow)
        self.processors: MutableSequence[CommandOutputProcessor] = processors
        self.check_processor: MutableMapping[Type[CommandOutputProcessor], Callable] = {
            CWLCommandOutputProcessor: _check_default_processor,
            CWLObjectCommandOutputProcessor: self._check_object_processor,
            CWLMapCommandOutputProcessor: self._check_map_processor,
            CWLUnionCommandOutputProcessor: self._check_union_processor}

    # noinspection PyMethodMayBeStatic
    def _check_map_processor(self, processor: CWLMapCommandOutputProcessor, token_value: Any):
        return isinstance(token_value, MutableSequence) and self.check_processor[type(processor.processor)](
            processor.processor, token_value[0])

    def _check_object_processor(self, processor: CWLObjectCommandOutputProcessor, token_value: Any):
        return (isinstance(token_value, MutableMapping) and
                all(k in processor.processors and
                    self.check_processor[type(processor.processors[k])](processor.processors[k], v)
                    for k, v in token_value.items()))

    def _check_union_processor(self, processor: CWLUnionCommandOutputProcessor, token_value: Any):
        return any(self.check_processor[type(p)](p, token_value) for p in processor.processors)

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'processors': await asyncio.gather(*(asyncio.create_task(v.save(context))
                                                        for v in self.processors))}}

    def get_processor(self, token_value: Any) -> CommandOutputProcessor:
        for processor in self.processors:
            if self.check_processor[type(processor)](processor, token_value):
                return processor
        raise WorkflowDefinitionException("No suitable token processors for value " + str(token_value))

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CommandOutputProcessor:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            processors=cast(MutableSequence[CommandOutputProcessor], await asyncio.gather(*(asyncio.create_task(
                CommandOutputProcessor.load(context, p, loading_context)) for p in row['processors']))))

    async def process(self,
                      job: Job,
                      command_output: CommandOutput,
                      connector: Optional[Connector] = None) -> Optional[Token]:
        token_value = command_output.value
        # If `token_value` is a dictionary, directly extract the token value from it
        if isinstance(token_value, MutableMapping) and self.name in token_value:
            token_value = token_value[self.name]
        return await self.get_processor(token_value).process(job, command_output, connector)



from os import path

from setuptools import setup

from streamflow.version import VERSION

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="streamflow",
    version=VERSION,
    packages=[
        "streamflow",
        "streamflow.config",
        "streamflow.core",
        "streamflow.cwl",
        "streamflow.cwl.antlr",
        "streamflow.data",
        "streamflow.deployment",
        "streamflow.deployment.connector",
        "streamflow.persistence",
        "streamflow.recovery",
        "streamflow.scheduling",
        "streamflow.scheduling.policy",
        "streamflow.workflow"
    ],
    package_data={
        "streamflow.config": ["schemas/v1.0/*.json"],
        "streamflow.data": ["schemas/*.json"],
        "streamflow.deployment": ["schemas/*.json"],
        "streamflow.deployment.connector": ["schemas/*.json"],
        "streamflow.persistence": ["schemas/*.sql", "schemas/*.json"],
        "streamflow.recovery": ["schemas/*.json"],
        "streamflow.scheduling.policy": ["schemas/*.json"]
    },
    include_package_data=True,
    url="https://github.com/alpha-unito/streamflow",
    download_url="".join(["https://github.com/alpha-unito/streamflow/releases"]),
    author="Iacopo Colonnelli",
    author_email="iacopo.colonnelli@unito.it",
    description="StreamFlow framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "aiohttp",
        "aiosqlite",
        "antlr4-python3-runtime>=4.10",
        "asyncssh",
        "bcrypt",
        "cachetools",
        "cwltool",
        "importlib_metadata",
        "Jinja2",
        "jsonref",
        "jsonschema",
        "kubernetes_asyncio>=22.6.3",
        "rdflib"
    ],
    extras_require={
        "report": [
            "pandas",
            "plotly",
            "kaleido"
        ],
        "docs": [
            "sphinx",
            "sphinx-jsonschema",
            "sphinx-rtd-theme"
        ]
    },
    tests_require=[
        'pytest',
        'cwltest'
    ],
    python_requires=">=3.8, <4",
    entry_points={"console_scripts": [
        "streamflow=streamflow.main:run",
        "cwl-runner=streamflow.cwl.runner:run"
    ]},
    zip_safe=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Distributed Computing",
    ],
)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from streamflow.core.config import SchemaEntity
from streamflow.core.data import DataLocation

if TYPE_CHECKING:
    from streamflow.core.context import StreamFlowContext
    from streamflow.core.workflow import Job, CommandOutput, Step
    from typing import Optional, MutableMapping, Any


class CheckpointManager(SchemaEntity):

    def __init__(self,
                 context: StreamFlowContext):
        self.context: StreamFlowContext = context

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    def register(self, data_location: DataLocation) -> None:
        ...


class FailureManager(SchemaEntity):

    def __init__(self, context: StreamFlowContext):
        self.context: StreamFlowContext = context

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    async def handle_exception(self,
                               job: Job,
                               step: Step,
                               exception: BaseException) -> CommandOutput:
        ...

    @abstractmethod
    async def handle_failure(self,
                             job: Job,
                             step: Step,
                             command_output: CommandOutput) -> CommandOutput:
        ...


class ReplayRequest(object):
    __slots__ = ('sender', 'target', 'version')

    def __init__(self, sender: str, target: str, version: int = 1):
        self.sender: str = sender
        self.target: str = target
        self.version: int = version


class ReplayResponse(object):
    __slots__ = ('job', 'outputs', 'version')

    def __init__(self, job: str, outputs: Optional[MutableMapping[str, Any]], version: int = 1):
        self.job: str = job
        self.outputs: MutableMapping[str, Any] = outputs
        self.version: int = version



import asyncio
from typing import Any, MutableSequence

from streamflow.core.context import StreamFlowContext
from streamflow.core.data import DataType
from streamflow.core.workflow import Token
from streamflow.cwl import utils
from streamflow.data import remotepath
from streamflow.workflow.token import FileToken


async def _get_file_token_weight(context: StreamFlowContext,
                                 value: Any):
    weight = 0
    if 'size' in value:
        weight = value['size']
    else:
        if path := utils.get_path_from_token(value):
            data_locations = context.data_manager.get_data_locations(
                path=path,
                location_type=DataType.PRIMARY)
            if data_locations:
                location = list(data_locations)[0]
                connector = context.deployment_manager.get_connector(location.deployment)
                real_path = await remotepath.follow_symlink(connector, location.location, location.path)
                weight = await remotepath.size(connector, location.location, real_path)
    if 'secondaryFiles' in value:
        weight += sum(await asyncio.gather(*(asyncio.create_task(
            _get_file_token_weight(
                context=context,
                value=sf
            )) for sf in value['secondaryFiles'])))
    return weight


async def _is_file_token_available(context: StreamFlowContext,
                                   value: Any) -> bool:
    if path := utils.get_path_from_token(value):
        data_locations = context.data_manager.get_data_locations(
            path=path,
            location_type=DataType.PRIMARY)
        return len(data_locations) != 0
    else:
        return True


class CWLFileToken(FileToken):

    async def get_paths(self, context: StreamFlowContext) -> MutableSequence[str]:
        paths = []
        if isinstance(self.value, MutableSequence):
            for value in self.value:
                if path := utils.get_path_from_token(value):
                    paths.append(path)
                for sf in value.get('secondaryFiles', []):
                    if path := utils.get_path_from_token(sf):
                        paths.append(path)
        elif self.value and (path := utils.get_path_from_token(self.value)):
            paths.append(path)
            for sf in self.value.get('secondaryFiles', []):
                if path := utils.get_path_from_token(sf):
                    paths.append(path)
        return paths

    async def get_weight(self, context):
        if isinstance(self.value, Token):
            return await self.value.get_weight(context)
        else:
            return await _get_file_token_weight(context, self.value)

    async def is_available(self, context: StreamFlowContext):
        if isinstance(self.value, Token):
            return await self.value.is_available(context)
        else:
            return await _is_file_token_available(context, self.value)



import logging

logger = logging.getLogger("streamflow")
defaultStreamHandler = logging.StreamHandler()
formatter = logging.Formatter(fmt='%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s',
                              datefmt='%Y-%m-%d %H:%M:%S')
defaultStreamHandler.setFormatter(formatter)
logger.addHandler(defaultStreamHandler)
logger.setLevel(logging.INFO)
logger.propagate = False



from __future__ import annotations

from pathlib import PurePosixPath
from typing import MutableSequence, TYPE_CHECKING

from streamflow.core import utils
from streamflow.core.config import Config
from streamflow.core.exception import WorkflowDefinitionException
from streamflow.log_handler import logger

if TYPE_CHECKING:
    from typing import MutableMapping, Any, Optional


def set_targets(current_node, target):
    for node in current_node['children'].values():
        if 'step' not in node:
            node['step'] = target
        set_targets(node, node['step'])


class WorkflowConfig(object):

    def __init__(self,
                 workflow_name: str,
                 streamflow_config: Optional[MutableMapping[str, Any]] = None) -> None:
        super().__init__()
        workflow_config = streamflow_config['workflows'][workflow_name]
        self.type = workflow_config['type']
        self.config = workflow_config['config']
        self.deplyoments = streamflow_config.get('deployments', {})
        self.policies = {k: Config(name=k, type=v['type'], config=v['config'])
                         for k, v in streamflow_config.get('scheduling', {}).get('policies', {}).items()}
        self.policies['__DEFAULT__'] = Config(name='__DEFAULT__', type='data_locality', config={})
        if not self.deplyoments:
            self.deplyoments = streamflow_config.get('models', {})
            if self.deplyoments:
                logger.warn(
                    "The `models` keyword is deprecated and will be removed in StreamFlow 0.3.0. "
                    "Use `deployments` instead.")
        self.scheduling_groups: MutableMapping[str, MutableSequence[str]] = {}
        for name, deployment in self.deplyoments.items():
            deployment['name'] = name
        self.filesystem = {'children': {}}
        for binding in workflow_config.get('bindings', []):
            if isinstance(binding, MutableSequence):
                for b in binding:
                    self._process_binding(b)
                self.scheduling_groups[utils.random_name()] = binding
            else:
                self._process_binding(binding)
        set_targets(self.filesystem, None)

    def _build_config(self, path: PurePosixPath) -> MutableMapping[str, Any]:
        current_node = self.filesystem
        for part in path.parts:
            if part not in current_node['children']:
                current_node['children'][part] = {'children': {}}
            current_node = current_node['children'][part]
        return current_node

    def _process_binding(self, binding: MutableMapping[str, Any]):
        current_config = self._build_config(PurePosixPath(
            binding['step'] if 'step' in binding else binding['port']))
        policy = binding['target'].get(
            'policy', self.deplyoments[binding['target'].get('deployment', binding['target'].get('model', {}))].get(
                'policy', '__DEFAULT__'))
        if policy not in self.policies:
            raise WorkflowDefinitionException("Policy {} is not defined".format(policy))
        binding['target']['policy'] = self.policies[policy]
        target_type = 'step' if 'step' in binding else 'port'
        if target_type == 'port' and 'workdir' not in binding['target']:
            raise WorkflowDefinitionException(
                "The `workdir` option is mandatory when specifying a `port` target.")
        current_config[target_type] = binding['target']

    def get(self, path: PurePosixPath, name: str, default: Optional[Any] = None) -> Optional[Any]:
        current_node = self.filesystem
        for part in path.parts:
            if part not in current_node['children']:
                return default
            current_node = current_node['children'][part]
        return current_node.get(name)

    def propagate(self, path: PurePosixPath, name: str) -> Optional[Any]:
        current_node = self.filesystem
        value = None
        for part in path.parts:
            if part not in current_node['children']:
                return value
            current_node = current_node['children'][part]
            if name in current_node:
                value = current_node[name]
        return value



import json
import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Mapping, MutableMapping, MutableSequence, Optional, Set

import antlr4
import cwltool.sandboxjs
from cwltool.sandboxjs import JavascriptException
from schema_salad.utils import json_dumps

from streamflow.core.exception import WorkflowDefinitionException, WorkflowExecutionException
from streamflow.core.utils import NamesStack
from streamflow.cwl.antlr.ECMAScriptLexer import ECMAScriptLexer
from streamflow.cwl.antlr.ECMAScriptListener import ECMAScriptListener
from streamflow.cwl.antlr.ECMAScriptParser import ECMAScriptParser


class _State(Enum):
    DEFAULT = 0
    DOLLAR = 1
    PAREN = 2
    BRACE = 3
    SINGLE_QUOTE = 4
    DOUBLE_QUOTE = 5
    ESCAPE = 6


_SEG_SYMBOL_STR = r"""\w+"""
_SEG_SINGLE_STR = r"""\['([^']|\\')+'\]"""
_SEG_DOUBLE_STR = r"""\["([^"]|\\")+"\]"""
_SEG_INDEX_STR = r"""\[[0-9]+\]"""
_SEGMENTS_STR = fr"(\.{_SEG_SYMBOL_STR}|{_SEG_SINGLE_STR}|{_SEG_DOUBLE_STR}|{_SEG_INDEX_STR})"
_REGEX_SEGMENT = re.compile(_SEGMENTS_STR, flags=re.UNICODE)
_PARAM_STR = fr"\(({_SEG_SYMBOL_STR}){_SEGMENTS_STR}*\)$"
_REGEX_PARAM = re.compile(_PARAM_STR, flags=re.UNICODE)


def _extract_key(next_seg: str) -> str:
    if next_seg[0] == ".":
        return next_seg[1:]
    elif next_seg[1] in ("'", '"'):
        return next_seg[2:-2].replace("\\'", "'").replace('\\"', '"')


class CWLDependencyListener(ECMAScriptListener):

    def __init__(self):
        self.deps: Set[str] = set()
        self.names: NamesStack = NamesStack()
        self.names.add_name("inputs")

    @staticmethod
    def _get_index(ctx: antlr4.ParserRuleContext) -> Optional[str]:
        token = ctx.getToken(ECMAScriptParser.StringLiteral, 0)
        return token.symbol.text if token else None

    @staticmethod
    def _get_name(ctx: antlr4.ParserRuleContext) -> Optional[str]:
        token = ctx.getToken(ECMAScriptParser.Identifier, 0)
        return token.symbol.text if token else None

    def enterFunctionDeclaration(self, ctx: ECMAScriptParser.FunctionDeclarationContext):
        self.names.add_scope()
        parameters = ctx.formalParameterList()
        if parameters:
            for param in parameters.Identifier():
                if (name := param.symbol.text) in self.names:
                    self.names.add_name(name)

    def exitFunctionDeclaration(self, ctx: ECMAScriptParser.FunctionDeclarationContext):
        self.names.delete_scope()

    def enterAssignmentExpression(self, ctx: ECMAScriptParser.AssignmentExpressionContext):
        left = ctx.getChild(0)
        right = ctx.getChild(2)
        if isinstance(left, ECMAScriptParser.SingleExpressionContext):
            left_name = self._get_name(left)
            if left_name:
                if left_name in self.names:
                    if isinstance(right, ECMAScriptParser.SingleExpressionContext):
                        right_name = self._get_name(right)
                        if right_name and right_name not in self.names:
                            self.names.delete_name(left_name)
                    else:
                        self.names.delete_name(left_name)
                elif isinstance(right, ECMAScriptParser.SingleExpressionContext):
                    right_name = self._get_name(right)
                    if right_name in self.names:
                        self.names.add_name(left_name)

    def enterMemberDotExpression(self, ctx: ECMAScriptParser.MemberDotExpressionContext):
        if self._get_name(ctx.singleExpression()) in self.names.global_names():
            if dep := self._get_name(ctx.identifierName()):
                self.deps.add(dep)

    def enterMemberIndexExpression(self, ctx: ECMAScriptParser.MemberIndexExpressionContext):
        if self._get_name(ctx.singleExpression()) in self.names.global_names():
            for expr in ctx.expressionSequence().singleExpression():
                if dep := self._get_index(expr.literal()).strip("'\""):
                    self.deps.add(dep)


class Evaluator(ABC):

    def __init__(self,
                 jslib: str,
                 rootvars: MutableMapping[str, Any],
                 timeout: float,
                 full_js: bool = False):
        self.jslib: str = jslib
        self.rootvars: MutableMapping[str, Any] = rootvars
        self.timeout: float = timeout
        self.full_js: bool = full_js

    @abstractmethod
    async def _ecmascript_evaluate(self, expr: str) -> Optional[Any]:
        ...

    @abstractmethod
    async def _regex_evaluate(self,
                              parsed_string: str,
                              remaining_string: str,
                              current_value: Any) -> Optional[Any]:
        ...

    def evaluate(self, expr: str) -> Optional[Any]:
        expression_parse_exception = None
        if (match := _REGEX_PARAM.match(expr)) is not None:
            first_symbol = match.group(1)
            first_symbol_end = match.end(1)
            if first_symbol_end + 1 == len(expr) and first_symbol == "null":
                return None
            try:
                if first_symbol not in self.rootvars:
                    raise WorkflowExecutionException("{} is not defined".format(first_symbol))
                return self._regex_evaluate(first_symbol, expr[first_symbol_end:-1], self.rootvars[first_symbol])

            except WorkflowExecutionException as wexc:
                expression_parse_exception = wexc
        if self.full_js:
            return self._ecmascript_evaluate(expr)
        else:
            if expression_parse_exception is not None:
                raise JavascriptException(
                    "Syntax error in parameter reference '{}': {}. This could be "
                    "due to using Javascript code without specifying "
                    "InlineJavascriptRequirement.".format(expr[1:-1], expression_parse_exception))
            else:
                raise JavascriptException(
                    "Syntax error in parameter reference '{}'. This could be due "
                    "to using Javascript code without specifying "
                    "InlineJavascriptRequirement.".format(expr))


class DependencyResolver(Evaluator):

    def _ecmascript_evaluate(self, expr: str) -> Optional[Any]:
        code = cwltool.sandboxjs.code_fragment_to_js(expr, self.jslib)
        lexer = ECMAScriptLexer(antlr4.InputStream(code))
        parser = ECMAScriptParser(antlr4.CommonTokenStream(lexer))
        listener = CWLDependencyListener()
        walker = antlr4.ParseTreeWalker()
        walker.walk(listener, parser.program())
        return listener.deps

    def _regex_evaluate(self,
                        parsed_string: str,
                        remaining_string: str,
                        current_value: Any) -> Optional[Any]:
        if parsed_string != "inputs":
            return set()
        elif remaining_string:
            if not (m := _REGEX_SEGMENT.match(remaining_string)):
                return set()
            key = _extract_key(m.group(1))
            if isinstance(current_value, MutableSequence) and key == "length" and not remaining_string[m.end(1):]:
                return set()
            return {key}
        else:
            return set()


class RuntimeEvaluator(Evaluator):

    def _ecmascript_evaluate(self, expr: str) -> Optional[Any]:
        return cwltool.sandboxjs.execjs(expr, self.jslib, self.timeout)

    def _regex_evaluate(self,
                        parsed_string: str,
                        remaining_string: str,
                        current_value: Any) -> Optional[Any]:
        if remaining_string:
            if not (m := _REGEX_SEGMENT.match(remaining_string)):
                return current_value
            next_seg = m.group(1)
            key = _extract_key(next_seg)
            if key is not None:
                if isinstance(current_value, MutableSequence) and key == "length" and not remaining_string[m.end(1):]:
                    return len(current_value)
                if not isinstance(current_value, MutableMapping):
                    raise WorkflowExecutionException("{} is a {}, cannot index on string '{}'".format(
                        parsed_string, type(current_value).__name__, key))
                if key not in current_value:
                    raise WorkflowExecutionException("{} does not contain key '{}'".format(
                        parsed_string, key))
            else:
                try:
                    key = int(next_seg[1:-1])
                except ValueError as v:
                    raise WorkflowExecutionException(str(v)) from v
                if not isinstance(current_value, MutableSequence):
                    raise WorkflowExecutionException("{} is a {}, cannot index on int '{}'".format(
                        parsed_string, type(current_value).__name__, key))
                if key and key >= len(current_value):
                    raise WorkflowExecutionException("{} list index {} out of range".format(
                        parsed_string, key))
            if (isinstance(current_value, Mapping) or
                    (isinstance(current_value, MutableSequence) and isinstance(key, int))):
                try:
                    return self._regex_evaluate(
                        ''.join([parsed_string, remaining_string]),
                        remaining_string[m.end(1):],
                        current_value[key])
                except KeyError:
                    raise WorkflowExecutionException("{} doesn't have property {}".format(
                        parsed_string, key))
            else:
                raise WorkflowExecutionException("{} doesn't have property {}".format(
                    parsed_string, key))
        else:
            return current_value


class Scanner(object):

    def __init__(self, text: str):
        self.text: str = text
        self.stack = [_State.DEFAULT]
        self.start = 0
        self.end = 0
        self.index = 0
        self.parts = []

    def _scan_container(self, c: str, open_char: str, close_char: str, state: _State) -> Optional[str]:
        if c == open_char:
            self.stack.append(state)
        elif c == close_char:
            self.stack.pop()
            if self.stack[-1] == _State.DOLLAR:
                self.stack = [_State.DEFAULT]
                if self.end < self.start - 1:
                    self.parts.append(self.text[self.end:self.start - 1])
                self.end = self.index
                return self.text[self.start:self.index]
        elif c == "'":
            self.stack.append(_State.SINGLE_QUOTE)
        elif c == '"':
            self.stack.append(_State.DOUBLE_QUOTE)

    def _scan_brace(self, c: str) -> Optional[str]:
        return self._scan_container(c, "{", "}", _State.BRACE)

    def _scan_default(self, c: str) -> None:
        if c == "$":
            self.stack.append(_State.DOLLAR)
        elif c == "\\":
            self.stack.append(_State.ESCAPE)

    def _scan_dollar(self, c: str) -> None:
        if c == "(":
            self.start = self.index - 1
            self.stack.append(_State.PAREN)
        elif c == "{":
            self.start = self.index - 1
            self.stack.append(_State.BRACE)
        else:
            self.stack.pop()

    def _scan_double_quote(self, c: str) -> None:
        self._scan_quote(c, '"')

    def _scan_escape(self, c: str) -> None:
        self.stack.pop()
        if self.stack[-1] == _State.DEFAULT:
            if c == "\\" or (c == "$" and self.text[self.index] in ["(", "{"]):
                if self.end < self.index - 2:
                    self.parts.append(self.text[self.end:self.index - 2])
                self.end = self.index - 1

    def _scan_paren(self, c: str) -> Optional[str]:
        return self._scan_container(c, "(", ")", _State.PAREN)

    def _scan_quote(self, c: str, quote_char: str):
        if c == quote_char:
            self.stack.pop()
        elif c == "\\":
            self.stack.append(_State.ESCAPE)

    def _scan_single_quote(self, c: str) -> None:
        self._scan_quote(c, "'")

    def add_part(self, part: str):
        self.parts.append(part)

    def get_result(self):
        return ''.join(self.parts)

    def scan(self) -> Optional[str]:
        while self.index < len(self.text):
            state = self.stack[-1]
            c = self.text[self.index]
            self.index += 1
            if state == _State.DEFAULT:
                self._scan_default(c)
            elif state == _State.ESCAPE:
                self._scan_escape(c)
            elif state == _State.DOLLAR:
                self._scan_dollar(c)
            elif state == _State.PAREN:
                if (expr := self._scan_paren(c)) is not None:
                    return expr
            elif state == _State.BRACE:
                if (expr := self._scan_brace(c)) is not None:
                    return expr
            elif state == _State.SINGLE_QUOTE:
                self._scan_single_quote(c)
            elif state == _State.DOUBLE_QUOTE:
                self._scan_double_quote(c)
        if len(self.stack) > 1 and not (len(self.stack) == 2 and self.stack[1] in (_State.ESCAPE, _State.DOLLAR)):
            raise WorkflowDefinitionException(
                "Substitution error, unfinished block starting at position {}: '{}' stack was {}".format(
                    self.start, self.text[self.start:], self.stack))
        else:
            if self.end < len(self.text) is not None:
                self.parts.append(self.text[self.end:])
            return None


def interpolate(
        text: str,
        rootvars: MutableMapping[str, Any],
        timeout: float = 600,
        full_js: bool = False,
        jslib: str = "",
        strip_whitespace: bool = True,
        resolve_dependencies: bool = False) -> Optional[Any]:
    if strip_whitespace:
        text = text.strip()
    scanner = Scanner(text)
    if resolve_dependencies:
        evaluator = DependencyResolver(
            jslib=jslib,
            rootvars=rootvars,
            timeout=timeout,
            full_js=full_js)
        dependencies = set()
        while (expr := scanner.scan()) is not None:
            dependencies.update(evaluator.evaluate(expr))
        return dependencies
    else:
        evaluator = RuntimeEvaluator(
            jslib=jslib,
            rootvars=rootvars,
            timeout=timeout,
            full_js=full_js)
        while (expr := scanner.scan()) is not None:
            result = evaluator.evaluate(expr)
            if not scanner.parts and scanner.index == len(text):
                return result
            leaf = json_dumps(result, sort_keys=True)
            if leaf[0] == '"':
                leaf = json.loads(leaf)
            scanner.add_part(leaf)
        return scanner.get_result()



import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import MutableMapping, MutableSequence, Optional

import pkg_resources
import psutil

from streamflow.core.context import StreamFlowContext
from streamflow.core.data import LOCAL_LOCATION
from streamflow.core.deployment import Connector
from streamflow.core.scheduling import Hardware, Location
from streamflow.deployment.connector.base import BaseConnector


def _get_disk_usage(path: Path):
    while not os.path.exists(path):
        path = path.parent
    return float(getattr(shutil.disk_usage(path), 'free') / 2 ** 20)


class LocalConnector(BaseConnector):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 transferBufferSize: int = 2 ** 16):
        super().__init__(deployment_name, context, transferBufferSize)
        self.cores = float(psutil.cpu_count())
        self.memory = float(psutil.virtual_memory().available / 2 ** 20)

    def _get_run_command(self,
                         command: str,
                         location: str,
                         interactive: bool = False):
        if sys.platform == 'win32':
            return "cmd /C '{command}'".format(command=command)
        else:
            return "sh -c '{command}'".format(command=command)

    async def _copy_remote_to_remote(self,
                                     src: str,
                                     dst: str,
                                     locations: MutableSequence[str],
                                     source_location: str,
                                     source_connector: Optional[Connector] = None,
                                     read_only: bool = False) -> None:
        source_connector = source_connector or self
        if source_connector == self:
            if os.path.isdir(src):
                os.makedirs(dst, exist_ok=True)
                shutil.copytree(src, dst, dirs_exist_ok=True)
            else:
                shutil.copy(src, dst)
        else:
            await super()._copy_remote_to_remote(
                src=src,
                dst=dst,
                locations=locations,
                source_connector=source_connector,
                source_location=source_location,
                read_only=read_only)

    async def deploy(self, external: bool) -> None:
        os.makedirs(os.path.join(tempfile.gettempdir(), 'streamflow'), exist_ok=True)

    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        return {LOCAL_LOCATION: Location(
            name=LOCAL_LOCATION,
            hostname='localhost',
            slots=1,
            hardware=Hardware(
                cores=self.cores,
                memory=self.memory,
                input_directory=_get_disk_usage(Path(input_directory)) if input_directory else float('inf'),
                output_directory=_get_disk_usage(Path(output_directory)) if output_directory else float('inf'),
                tmp_directory=_get_disk_usage(Path(tmp_directory)) if tmp_directory else float('inf')))}

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'local.json'))

    async def undeploy(self, external: bool) -> None:
        pass



from typing import Optional

from streamflow.core.deployment import Connector
from streamflow.core.workflow import Job, Port, Token
from streamflow.workflow.token import TerminationToken


class ConnectorPort(Port):

    async def get_connector(self, consumer: str) -> Connector:
        token = await self.get(consumer)
        return self.workflow.context.deployment_manager.get_connector(token.value)

    def put_connector(self, connector_name: str):
        self.put(Token(value=connector_name))


class JobPort(Port):

    async def get_job(self, consumer: str) -> Optional[Job]:
        token = await self.get(consumer)
        if isinstance(token, TerminationToken):
            return None
        else:
            return token.value

    def put_job(self, job: Job):
        self.put(Token(value=job))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import MutableMapping, Type

from streamflow.core.deployment import Connector
from streamflow.core.scheduling import Policy, Scheduler
from streamflow.deployment.connector import connector_classes
from streamflow.log_handler import logger
from streamflow.scheduling import scheduler_classes
from streamflow.scheduling.policy import policy_classes


class StreamFlowPlugin(ABC):

    def _register(self,
                  name: str,
                  cls: Type,
                  classes: MutableMapping[str, Type]):
        if name in classes:
            logger.warning("{} is already installed and will be overridden by {}".format(
                name, self.__class__.__module__ + "." + self.__class__.__name__))
        classes[name] = cls

    @abstractmethod
    def register(self) -> None:
        ...

    def register_connector(self,
                           name: str,
                           cls: Type[Connector]):
        self._register(name, cls, connector_classes)

    def register_policy(self,
                        name: str,
                        cls: Type[Policy]):
        self._register(name, cls, policy_classes)

    def register_scheduler(self,
                           cls: Type[Scheduler]):
        self._register(cls.__class__.__module__ + '.' + cls.__class__.__name__, cls, scheduler_classes)



"""
Helpers to use [cachetools](https://github.com/tkem/cachetools) with
asyncio.
"""
import functools
import inspect

__all__ = ["cached", "cachedmethod"]

import cachetools


# noinspection PyUnresolvedReferences
def cached(cache, key=cachetools.keys.hashkey, lock=None):
    """
    Decorator to wrap a function or a coroutine with a memoizing callable
    that saves results in a cache.
    When ``lock`` is provided for a standard function, it's expected to
    implement ``__enter__`` and ``__exit__`` that will be used to lock
    the cache when gets updated. If it wraps a coroutine, ``lock``
    must implement ``__aenter__`` and ``__aexit__``.
    """

    def decorator(func):
        if inspect.iscoroutinefunction(func):
            if cache is None:
                async def wrapper(*args, **kwargs):
                    return await func(*args, **kwargs)
            elif lock is None:
                async def wrapper(*args, **kwargs):
                    k = key(*args, **kwargs)
                    try:
                        return cache[k]
                    except KeyError:
                        pass  # key not found
                    v = await func(*args, **kwargs)
                    try:
                        cache[k] = v
                    except ValueError:
                        pass  # value too large
                    return v
            else:
                async def wrapper(*args, **kwargs):
                    k = key(*args, **kwargs)
                    try:
                        async with lock:
                            return cache[k]
                    except KeyError:
                        pass  # key not found
                    v = await func(*args, **kwargs)
                    # in case of a race, prefer the item already in the cache
                    try:
                        async with lock:
                            return cache.setdefault(k, v)
                    except ValueError:
                        return v  # value too large
            return functools.update_wrapper(wrapper, func)
        else:
            raise NotImplementedError("Use cachetools.cached for non-async functions")

    return decorator


# noinspection PyUnresolvedReferences
def cachedmethod(cache, key=cachetools.keys.hashkey, lock=None):
    """Decorator to wrap a class or instance method with a memoizing
    callable that saves results in a cache.
    When ``lock`` is provided for a standard function, it's expected to
    implement ``__enter__`` and ``__exit__`` that will be used to lock
    the cache when gets updated. If it wraps a coroutine, ``lock``
    must implement ``__aenter__`` and ``__aexit__``.
    """

    def decorator(method):
        if inspect.iscoroutinefunction(method):
            if lock is None:
                async def wrapper(self, *args, **kwargs):
                    c = cache(self)
                    if c is None:
                        return await method(self, *args, **kwargs)
                    k = key(*args, **kwargs)
                    try:
                        return c[k]
                    except KeyError:
                        pass  # key not found
                    v = await method(self, *args, **kwargs)
                    try:
                        c[k] = v
                    except ValueError:
                        pass  # value too large
                    return v
            else:
                async def wrapper(self, *args, **kwargs):
                    c = cache(self)
                    if c is None:
                        return await method(self, *args, **kwargs)
                    k = key(*args, **kwargs)
                    try:
                        async with lock(self):
                            return c[k]
                    except KeyError:
                        pass  # key not found
                    v = await method(self, *args, **kwargs)
                    # in case of a race, prefer the item already in the cache
                    try:
                        async with lock(self):
                            return c.setdefault(k, v)
                    except ValueError:
                        return v  # value too large
            return functools.update_wrapper(wrapper, method)
        else:
            raise NotImplementedError("Use cachetools.cachedmethod for non-async methods")

    return decorator



from __future__ import annotations

import asyncio
import base64
import json
import logging
import posixpath
import re
import shlex
import time
from abc import ABC
from asyncio.subprocess import STDOUT
from typing import Any, IO, Iterable, List, MutableMapping, MutableSequence, Optional, Union

from streamflow.core.data import DataLocation, LOCAL_LOCATION
from streamflow.core.deployment import Connector
from streamflow.core.exception import WorkflowDefinitionException, WorkflowExecutionException
from streamflow.core.utils import flatten_list, get_path_processor
from streamflow.core.workflow import Command, CommandOutput, Job, Status, Step, Token, Workflow
from streamflow.cwl import utils
from streamflow.data import remotepath
from streamflow.log_handler import logger


def _adjust_inputs(inputs: MutableSequence[MutableMapping[str, Any]],
                   path_processor,
                   src_path: str,
                   dest_path: str) -> MutableSequence[MutableMapping[str, Any]]:
    for inp in inputs:
        if (token_class := utils.get_token_class(inp)) in ('File', 'Directory'):
            path = utils.get_path_from_token(inp)
            if path == src_path:
                inp['path'] = dest_path
                dirname, basename = path_processor.split(dest_path)
                inp['dirname'] = dirname
                inp['basename'] = basename
                if token_class == 'File':
                    nameroot, nameext = path_processor.splitext(basename)
                    inp['nameroot'] = nameroot
                    inp['nameext'] = nameext
            elif token_class == 'Directory' and src_path.startswith(path) and 'listing' in inp:
                inp['listing'] = _adjust_inputs(inp['listing'], path_processor, src_path, dest_path)
    return inputs


def _check_command_token(command_token: CWLCommandToken, input_value: Any) -> bool:
    # CWLMapCommandToken is suitable for input lists
    if isinstance(command_token, CWLMapCommandToken):
        if isinstance(input_value, MutableSequence):
            return _check_command_token(command_token.value, input_value[0])
        else:
            return False
    # At least one command token in a CWLUnionCommandToken must be suitable for the input value
    elif isinstance(command_token, CWLUnionCommandToken):
        for ct in command_token.value:
            if _check_command_token(ct, input_value):
                return True
        return False
    # A CWLObjectCommandToken must match the input value structure to be suitable
    elif isinstance(command_token, CWLObjectCommandToken):
        if isinstance(input_value, MutableMapping):
            for k, v in input_value.items():
                if not (k in command_token.value and _check_command_token(command_token.value[k], v)):
                    return False
            return True
        else:
            return False
    # A default command token must match its type with the token value type
    else:
        if isinstance(input_value, Token):
            input_value = input_value.value
        return command_token.token_type == utils.infer_type_from_token(input_value)


async def _check_cwl_output(job: Job, step: Step, result: Any) -> Any:
    connector = step.workflow.context.scheduler.get_connector(job.name)
    locations = step.workflow.context.scheduler.get_locations(job.name)
    path_processor = get_path_processor(connector)
    cwl_output_path = path_processor.join(job.output_directory, 'cwl.output.json')
    for location in locations:
        if await remotepath.exists(connector, location, cwl_output_path):
            # If file exists, use its contents as token value
            result = json.loads(await remotepath.read(connector, location, cwl_output_path))
            # Update step output ports at runtime if needed
            if isinstance(result, MutableMapping):
                for out_name, out in result.items():
                    if out_name not in step.output_ports:
                        step.add_output_port(
                            out_name,
                            step.workflow.create_port())
                        # Update workflow outputs if needed
                        if out_name not in step.workflow.output_ports:
                            step.workflow.output_ports[out_name] = step.output_ports[out_name]
                for out_name in [p for p in step.output_ports if p not in result]:
                    result[out_name] = None
                    if out_name in step.workflow.output_ports:
                        del step.workflow.output_ports[out_name]
            break
    return result


async def _get_source_location(src_path: str,
                               input_directory: str,
                               connector: Connector,
                               workflow: Workflow) -> Optional[DataLocation]:
    path_processor = get_path_processor(connector)
    src_path = path_processor.normpath(src_path)
    input_directory = path_processor.normpath(input_directory)
    relpath = path_processor.basename(src_path)
    base_path = None
    # If source path refers to job input directory
    if src_path.startswith(input_directory):
        relpath = path_processor.join(*path_processor.relpath(src_path, input_directory).split(path_processor.sep)[1:])
        base_path = path_processor.normpath(src_path[:-len(relpath)])
    # Register parent locations
    await utils.search_in_parent_locations(
        context=workflow.context,
        connector=connector,
        path=src_path,
        relpath=relpath,
        base_path=base_path)
    # Return the best source location ofr the current transfer
    return workflow.context.data_manager.get_source_location(src_path, connector.deployment_name)


def _get_value_for_command(token: Any, item_separator: Optional[str]) -> Any:
    if isinstance(token, MutableSequence):
        value = []
        for element in token:
            value.append(_get_value_for_command(element, item_separator))
        if item_separator is not None:
            value = item_separator.join([str(v) for v in value])
        return value or None
    elif isinstance(token, MutableMapping):
        if (path := utils.get_path_from_token(token)) is not None:
            return path
        else:
            raise WorkflowExecutionException("Unsupported value " + str(token) + " from expression")
    else:
        return token


def _escape_value(value: Any) -> Any:
    if isinstance(value, MutableSequence):
        return [_escape_value(v) for v in value]
    else:
        return shlex.quote(str(value))


def alphanumeric_sorting(list_to_sort: Iterable) -> Iterable:
    regex = '(-{0,1}[0-9]+)'
    return sorted(list_to_sort, key=lambda key: [int(c) if re.match(regex, c) else c for c in re.split(regex, key)])


def _merge_tokens(bindings_map: MutableMapping[str, MutableSequence[Any]]) -> MutableSequence[Any]:
    command = []
    for binding_position in alphanumeric_sorting(bindings_map.keys()):
        for binding in bindings_map[binding_position]:
            command.extend(flatten_list(binding))
    return [str(token) for token in command]


class CWLBaseCommand(Command, ABC):

    def __init__(self,
                 step: Step,
                 absolute_initial_workdir_allowed: bool = False,
                 expression_lib: Optional[MutableSequence[str]] = None,
                 initial_work_dir: Optional[Union[str, MutableSequence[Any]]] = None,
                 inplace_update: bool = False,
                 full_js: bool = False,
                 time_limit: Optional[Union[int, str]] = None):
        super().__init__(step)
        self.absolute_initial_workdir_allowed: bool = absolute_initial_workdir_allowed
        self.expression_lib: Optional[MutableSequence[str]] = expression_lib
        self.full_js: bool = full_js
        self.initial_work_dir: Optional[Union[str, MutableSequence[Any]]] = initial_work_dir
        self.inplace_update: bool = inplace_update
        self.time_limit: Optional[Union[int, str]] = time_limit

    def _get_timeout(self, job: Job) -> Optional[int]:
        if isinstance(self.time_limit, int):
            timeout = self.time_limit
        elif isinstance(self.time_limit, str):
            context = utils.build_context(
                inputs=job.inputs,
                output_directory=job.output_directory,
                tmp_directory=job.tmp_directory,
                hardware=self.step.workflow.context.scheduler.get_hardware(job.name))
            timeout = int(utils.eval_expression(
                expression=self.time_limit,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib))
        else:
            timeout = 0
        return timeout if timeout > 0 else None

    async def _prepare_work_dir(self,
                                job: Job,
                                context: MutableMapping[str, Any],
                                element: Any,
                                base_path: Optional[str] = None,
                                dest_path: Optional[str] = None,
                                writable: bool = False) -> None:
        connector = self.step.workflow.context.scheduler.get_connector(job.name)
        locations = self.step.workflow.context.scheduler.get_locations(job.name)
        path_processor = get_path_processor(connector)
        # Initialize base path to job output directory if present
        base_path = base_path or job.output_directory
        # If current element is a string, it must be an expression
        if isinstance(element, str):
            listing = utils.eval_expression(
                expression=element,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib)
        else:
            listing = element
        # If listing is a list, each of its elements must be processed independently
        if isinstance(listing, MutableSequence):
            await asyncio.gather(*(asyncio.create_task(
                self._prepare_work_dir(
                    job=job,
                    context=context,
                    element=el,
                    base_path=base_path,
                    writable=writable)
            ) for el in listing))
        # If listing is a dictionary, it could be a File, a Directory, a Dirent or some other object
        elif isinstance(listing, MutableMapping):
            # If it is a File or Directory element, put the correspnding file in the output directory
            if (listing_class := utils.get_token_class(listing)) in ['File', 'Directory']:
                src_path = utils.get_path_from_token(listing)
                # If a compatible source location exists, simply transfer data
                if (src_path is not None and (selected_location := await _get_source_location(
                        src_path, job.input_directory, connector, self.step.workflow))):
                    dest_path = dest_path or path_processor.join(base_path, path_processor.basename(src_path))
                    await self.step.workflow.context.data_manager.transfer_data(
                        src_deployment=selected_location.deployment,
                        src_locations=[selected_location.location],
                        src_path=src_path,
                        dst_deployment=connector.deployment_name,
                        dst_locations=locations,
                        dst_path=dest_path,
                        writable=writable)
                    _adjust_inputs(
                        inputs=context['inputs'].values(),
                        path_processor=path_processor,
                        src_path=src_path,
                        dest_path=dest_path)
                # Otherwise create a File or a Directory in the remote path
                else:
                    if dest_path is None:
                        dest_path = base_path
                    if src_path is not None:
                        dest_path = path_processor.join(dest_path, path_processor.basename(src_path))
                    if listing_class == 'Directory':
                        await remotepath.mkdir(connector, locations, dest_path)
                    else:
                        await utils.write_remote_file(
                            context=self.step.workflow.context,
                            job=job,
                            content=listing['contents'] if 'contents' in listing else '',
                            path=dest_path)
                # If `listing` is present, recursively process folder contents
                if 'listing' in listing:
                    if 'basename' in listing:
                        folder_path = path_processor.join(base_path, listing['basename'])
                        await remotepath.mkdir(connector, locations, folder_path)
                    else:
                        folder_path = dest_path or base_path
                    await asyncio.gather(*(asyncio.create_task(
                        self._prepare_work_dir(
                            job=job,
                            context=context,
                            element=element,
                            base_path=folder_path,
                            writable=writable)
                    ) for element in listing['listing']))
                # If `secondaryFiles` is present, resursively process secondary files
                if 'secondaryFiles' in listing:
                    await asyncio.gather(*(asyncio.create_task(
                        self._prepare_work_dir(
                            job=job,
                            context=context,
                            element=element,
                            base_path=base_path,
                            writable=writable)
                    ) for element in listing['secondaryFiles']))
            # If it is a Dirent element, put or create the corresponding file according to the entryname field
            elif 'entry' in listing:
                entry = utils.eval_expression(
                    expression=listing['entry'],
                    context=context,
                    full_js=self.full_js,
                    expression_lib=self.expression_lib,
                    strip_whitespace=False)
                if 'entryname' in listing:
                    dest_path = utils.eval_expression(
                        expression=listing['entryname'],
                        context=context,
                        full_js=self.full_js,
                        expression_lib=self.expression_lib)
                    if not path_processor.isabs(dest_path):
                        dest_path = posixpath.abspath(posixpath.join(job.output_directory, dest_path))
                    if not (dest_path.startswith(job.output_directory) or self.absolute_initial_workdir_allowed):
                        raise WorkflowDefinitionException("Entryname cannot be outside the job's output directory")
                    # The entryname field overrides the value of basename of the File or Directory object
                    if (isinstance(entry, MutableMapping) and
                            'path' not in entry and 'location' not in entry and
                            'basename' in entry):
                        entry['basename'] = dest_path
                    if not path_processor.isabs(dest_path):
                        dest_path = path_processor.join(job.output_directory, dest_path)
                writable = listing['writable'] if 'writable' in listing and not self.inplace_update else False
                # If entry is a string, a new text file must be created with the string as the file contents
                if isinstance(entry, str):
                    await utils.write_remote_file(
                        context=self.step.workflow.context,
                        job=job,
                        content=entry,
                        path=dest_path or base_path)
                # If entry is a list
                elif isinstance(entry, MutableSequence):
                    # If all elements are Files or Directories, each of them must be processed independently
                    if all(utils.get_token_class(t) in ['File', 'Directory'] for t in entry):
                        await self._prepare_work_dir(
                            job=job,
                            context=context,
                            element=entry,
                            base_path=base_path,
                            dest_path=dest_path,
                            writable=writable)
                    # Otherwise, the content should be serialised to JSON
                    else:
                        await utils.write_remote_file(
                            context=self.step.workflow.context,
                            job=job,
                            content=json.dumps(entry),
                            path=dest_path or base_path)
                # If entry is a dict
                elif isinstance(entry, MutableMapping):
                    # If it is a File or Directory, it must be put in the destination path
                    if utils.get_token_class(entry) in ['File', 'Directory']:
                        await self._prepare_work_dir(
                            job=job,
                            context=context,
                            element=entry,
                            base_path=base_path,
                            dest_path=dest_path,
                            writable=writable)
                    # Otherwise, the content should be serialised to JSON
                    else:
                        await utils.write_remote_file(
                            context=self.step.workflow.context,
                            job=job,
                            content=json.dumps(entry),
                            path=dest_path or base_path)
                # Every object different from a string should be serialised to JSON
                elif entry is not None:
                    await utils.write_remote_file(
                        context=self.step.workflow.context,
                        job=job,
                        content=json.dumps(entry),
                        path=dest_path or base_path)


class CWLCommand(CWLBaseCommand):

    def __init__(self,
                 step: Step,
                 absolute_initial_workdir_allowed: bool = False,
                 expression_lib: Optional[MutableSequence[str]] = None,
                 failure_codes: Optional[MutableSequence[int]] = None,
                 full_js: bool = False,
                 initial_work_dir: Optional[Union[str, MutableSequence[Any]]] = None,
                 inplace_update: bool = False,
                 is_shell_command: bool = False,
                 success_codes: Optional[MutableSequence[int]] = None,
                 step_stderr: Optional[str] = None,
                 step_stdin: Optional[str] = None,
                 step_stdout: Optional[str] = None,
                 time_limit: Optional[Union[int, str]] = None):
        super().__init__(
            step=step,
            absolute_initial_workdir_allowed=absolute_initial_workdir_allowed,
            expression_lib=expression_lib,
            initial_work_dir=initial_work_dir,
            inplace_update=inplace_update,
            full_js=full_js,
            time_limit=time_limit)
        self.base_command: MutableSequence[str] = []
        self.command_tokens: MutableSequence[CWLCommandToken] = []
        self.environment: MutableMapping[str, str] = {}
        self.failure_codes: Optional[MutableSequence[int]] = failure_codes
        self.is_shell_command: bool = is_shell_command
        self.success_codes: Optional[MutableSequence[int]] = success_codes
        self.stderr: Union[str, IO] = step_stderr
        self.stdin: Union[str, IO] = step_stdin
        self.stdout: Union[str, IO] = step_stdout

    def _get_executable_command(self, context: MutableMapping[str, Any]) -> MutableSequence[str]:
        command = []
        bindings_map = {}
        # Process baseCommand
        if self.base_command:
            command.append(shlex.join(self.base_command))
        # Process tokens
        for command_token in self.command_tokens:
            if command_token.name is not None:
                context['self'] = context['inputs'].get(command_token.name)
                # If input is None, skip the command token
                if context['self'] is None:
                    continue
            bindings_map = command_token.get_binding(
                context=context,
                bindings_map=bindings_map,
                full_js=self.full_js,
                expression_lib=self.expression_lib)
        # Merge tokens
        command.extend(_merge_tokens(bindings_map))
        return command

    async def execute(self, job: Job) -> CWLCommandOutput:
        # Build context
        context = utils.build_context(
            inputs=job.inputs,
            output_directory=job.output_directory,
            tmp_directory=job.tmp_directory,
            hardware=self.step.workflow.context.scheduler.get_hardware(job.name))
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Job {job} inputs: {inputs}".format(
                job=job.name, inputs=json.dumps(context['inputs'], indent=4, sort_keys=True)))
        # Build working directory
        if self.initial_work_dir is not None:
            await self._prepare_work_dir(job, context, self.initial_work_dir)
        # Build command string
        cmd = self._get_executable_command(context)
        # Build environment variables
        parsed_env = {k: str(utils.eval_expression(
            expression=v,
            context=context,
            full_js=self.full_js,
            expression_lib=self.expression_lib
        )) for (k, v) in self.environment.items()}
        if 'HOME' not in parsed_env:
            parsed_env['HOME'] = job.output_directory
        if 'TMPDIR' not in parsed_env:
            parsed_env['TMPDIR'] = job.tmp_directory
        # Get execution target
        connector = self.step.workflow.context.scheduler.get_connector(job.name)
        locations = self.step.workflow.context.scheduler.get_locations(job.name)
        cmd_string = ' \\\n\t'.join(["/bin/sh", "-c", "\"{cmd}\"".format(cmd=" ".join(cmd))]
                                    if self.is_shell_command else cmd)
        logger.info('Executing step {step} (job {job}) {location} into directory {outdir}:\n{command}'.format(
            step=self.step.name,
            job=job.name,
            location="locally" if locations[0] == LOCAL_LOCATION else "on location {loc}".format(
                loc=locations[0]),
            outdir=job.output_directory,
            command=cmd_string))
        # Persist command
        command_id = await self.step.workflow.context.database.add_command(
            step_id=self.step.persistent_id,
            cmd=cmd_string)
        # Escape shell command when needed
        if self.is_shell_command:
            cmd = ["/bin/sh", "-c", "\"$(echo {command} | base64 -d)\"".format(
                command=base64.b64encode(" ".join(cmd).encode('utf-8')).decode('utf-8'))]
        # If step is assigned to multiple locations, add the STREAMFLOW_HOSTS environment variable
        if len(locations) > 1:
            service = self.step.workflow.context.scheduler.get_service(job.name)
            available_locations = await connector.get_available_locations(
                service=service)
            hosts = {k: v.hostname for k, v in available_locations.items() if k in locations}
            parsed_env['STREAMFLOW_HOSTS'] = ','.join(hosts.values())
        # Process streams
        stdin = utils.eval_expression(
            expression=self.stdin,
            context=context,
            full_js=self.full_js,
            expression_lib=self.expression_lib)
        stdout = utils.eval_expression(
            expression=self.stdout,
            context=context,
            full_js=self.full_js,
            expression_lib=self.expression_lib
        ) if self.stdout is not None else STDOUT
        stderr = utils.eval_expression(
            expression=self.stderr,
            context=context,
            full_js=self.full_js,
            expression_lib=self.expression_lib
        ) if self.stderr is not None else stdout
        # Execute remote command
        start_time = time.time_ns()
        result, exit_code = await asyncio.wait_for(
            connector.run(
                locations[0] if locations else None,
                cmd,
                environment=parsed_env,
                workdir=job.output_directory,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                capture_output=True,
                job_name=job.name),
            self._get_timeout(job))
        end_time = time.time_ns()
        # Handle exit codes
        if self.failure_codes and exit_code in self.failure_codes:
            status = Status.FAILED
        elif (self.success_codes and exit_code in self.success_codes) or exit_code == 0:
            status = Status.COMPLETED
            if result:
                logger.info(result)
        else:
            status = Status.FAILED
        # Update command persistence
        await self.step.workflow.context.database.update_command(command_id, {
            "status": status.value,
            "output": str(result),
            "start_time": start_time,
            "end_time": end_time})
        # Check if file `cwl.output.json` exists either locally on at least one location
        result = await _check_cwl_output(job, self.step, result)
        return CWLCommandOutput(value=result, status=status, exit_code=exit_code)


class CWLCommandOutput(CommandOutput):
    __slots__ = 'exit_code'

    def __init__(self,
                 value: Any,
                 status: Status,
                 exit_code: int):
        super().__init__(value, status)
        self.exit_code: int = exit_code

    def update(self, value: Any):
        return CWLCommandOutput(value=value, status=self.status, exit_code=self.exit_code)


class CWLCommandToken(object):

    def __init__(self,
                 name: str,
                 value: Any,
                 token_type: Optional[str] = None,
                 is_shell_command: bool = False,
                 item_separator: Optional[str] = None,
                 position: Union[str, int] = 0,
                 prefix: Optional[str] = None,
                 separate: bool = True,
                 shell_quote: bool = True):
        self.name: str = name
        self.value: Any = value
        self.token_type: Optional[str] = token_type
        self.is_shell_command: bool = is_shell_command
        self.item_separator: Optional[str] = item_separator
        self.position: Union[str, int] = position
        self.prefix: Optional[str] = prefix
        self.separate: bool = separate
        self.shell_quote: bool = shell_quote

    def _compute_binding(self,
                         processed_token: Any,
                         context: MutableMapping[str, Any],
                         bindings_map: MutableMapping[str, MutableSequence[Any]],
                         full_js: bool = False,
                         expression_lib: Optional[MutableSequence[str]] = None
                         ) -> MutableMapping[str, MutableSequence[Any]]:
        # Obtain token value
        value = _get_value_for_command(processed_token, self.item_separator)
        # If token value is null or an empty array, skip the command token
        if value is None:
            return bindings_map
        # Otherwise
        else:
            # Obtain prefix if present
            if self.prefix is not None:
                if isinstance(value, bool):
                    value = [self.prefix] if value else value
                elif self.separate:
                    if isinstance(value, MutableSequence):
                        value = [self.prefix] + list(value)
                    else:
                        value = [self.prefix, value]
                elif isinstance(value, MutableSequence):
                    value = [self.prefix].extend(value)
                else:
                    value = [self.prefix + str(value)]
            # If value is a boolean with no prefix, skip it
            if isinstance(value, bool):
                return bindings_map
            # Ensure value is a list
            if not isinstance(value, MutableSequence):
                value = [value]
            # Process shell escape only on the single command token
            if not self.is_shell_command or self.shell_quote:
                value = [_escape_value(v) for v in value]
            # Obtain token position
            if isinstance(self.position, str) and not self.position.isnumeric():
                context['self'] = processed_token
                position = utils.eval_expression(
                    expression=self.position,
                    context=context,
                    full_js=full_js,
                    expression_lib=expression_lib)
                try:
                    position = int(position) if position is not None else 0
                except ValueError:
                    pass
            else:
                position = int(self.position)
            sort_key = "".join([str(position), self.name]) if self.name else str(position)
            if sort_key not in bindings_map:
                bindings_map[sort_key] = []
            # Place value in proper position
            bindings_map[sort_key].append(value)
            return bindings_map

    def _process_token(self,
                       token_value: Any,
                       context: MutableMapping[str, Any],
                       full_js: bool = False,
                       expression_lib: Optional[MutableSequence[str]] = None) -> Any:
        if isinstance(token_value, CWLCommandToken):
            local_bindings = token_value.get_binding(
                context=context,
                bindings_map={},
                full_js=full_js,
                expression_lib=expression_lib)
            processed_token = _merge_tokens(local_bindings)
        elif isinstance(token_value, str):
            processed_token = utils.eval_expression(
                expression=self.value,
                context=context,
                full_js=full_js,
                expression_lib=expression_lib)
        elif token_value is None:
            processed_token = context['inputs'][self.name]
            processed_token = processed_token.value if isinstance(processed_token, Token) else processed_token
        else:
            processed_token = token_value
        return processed_token

    def get_binding(self,
                    context: MutableMapping[str, Any],
                    bindings_map: MutableMapping[str, MutableSequence[Any]],
                    full_js: bool = False,
                    expression_lib: Optional[MutableSequence[str]] = None) -> Any:
        processed_token = self._process_token(
            token_value=self.value,
            context=context,
            full_js=full_js,
            expression_lib=expression_lib)
        return self._compute_binding(
            processed_token=processed_token,
            context=context,
            bindings_map=bindings_map,
            full_js=full_js,
            expression_lib=expression_lib)


class CWLObjectCommandToken(CWLCommandToken):

    def _check_dict(self, value: Any):
        if not isinstance(value, MutableMapping):
            raise WorkflowDefinitionException(
                "A {this} object can only be used to process dict values".format(this=self.__class__.__name__))

    def get_binding(self,
                    context: MutableMapping[str, Any],
                    bindings_map: MutableMapping[str, MutableSequence[Any]],
                    full_js: bool = False,
                    expression_lib: Optional[MutableSequence[str]] = None) -> Any:
        self._check_dict(self.value)
        context = {**context, **{'inputs': context['inputs'][self.name]}}
        for key, token in self.value.items():
            if key in context['inputs'] and token is not None:
                bindings_map = token.get_binding(
                    context=context,
                    bindings_map=bindings_map,
                    full_js=full_js,
                    expression_lib=expression_lib)
        return bindings_map


class CWLUnionCommandToken(CWLCommandToken):

    def get_binding(self,
                    context: MutableMapping[str, Any],
                    bindings_map: MutableMapping[str, MutableSequence[Any]],
                    full_js: bool = False,
                    expression_lib: Optional[MutableSequence[str]] = None) -> Any:
        inputs = context['inputs'][self.name]
        command_token = self.get_command_token(inputs)
        return command_token.get_binding(
            context=context,
            bindings_map=bindings_map,
            full_js=full_js,
            expression_lib=expression_lib)

    def get_command_token(self, inputs: Any) -> CWLCommandToken:
        if self.value is not None:
            for command_token in self.value:
                if _check_command_token(command_token, inputs):
                    return command_token
            raise WorkflowDefinitionException("No suitable command token for input value " + str(inputs))
        else:
            return super()


class CWLMapCommandToken(CWLCommandToken):

    def _check_list(self, value: Any):
        if not isinstance(value, MutableSequence):
            raise WorkflowDefinitionException(
                "A {this} object can only be used to process list values".format(this=self.__class__.__name__))

    def get_binding(self,
                    context: MutableMapping[str, Any],
                    bindings_map: MutableMapping[str, MutableSequence[Any]],
                    full_js: bool = False,
                    expression_lib: Optional[MutableSequence[str]] = None) -> Any:
        inputs = context['inputs'][self.name]
        if isinstance(inputs, Token):
            inputs = inputs.value
        self._check_list(inputs)
        for value in inputs:
            bindings_map = super().get_binding(
                context={**context, **{
                    'inputs': {self.name: value},
                    'self': value
                }},
                bindings_map=bindings_map,
                full_js=full_js,
                expression_lib=expression_lib)
        return bindings_map


class CWLExpressionCommand(CWLBaseCommand):

    def __init__(self,
                 step: Step,
                 expression: str,
                 absolute_initial_workdir_allowed: bool = False,
                 expression_lib: Optional[List[str]] = None,
                 full_js: bool = False,
                 initial_work_dir: Optional[Union[str, List[Any]]] = None,
                 inplace_update: bool = False):
        super().__init__(
            step=step,
            absolute_initial_workdir_allowed=absolute_initial_workdir_allowed,
            expression_lib=expression_lib,
            initial_work_dir=initial_work_dir,
            inplace_update=inplace_update,
            full_js=full_js)
        self.expression: str = expression

    async def execute(self, job: Job) -> CWLCommandOutput:
        context = utils.build_context(
            inputs=job.inputs,
            output_directory=job.output_directory,
            tmp_directory=job.tmp_directory,
            hardware=self.step.workflow.context.scheduler.get_hardware(job.name))
        if self.initial_work_dir is not None:
            await self._prepare_work_dir(job, context, self.initial_work_dir)
        logger.info('Evaluating expression for step {step} (job {job})'.format(
            step=self.step.name, job=job.name))
        # Persist command
        command_id = await self.step.workflow.context.database.add_command(
            self.step.persistent_id,
            self.expression)
        # Execute command
        start_time = time.time_ns()
        result = utils.eval_expression(
            expression=self.expression,
            context=context,
            full_js=self.full_js,
            expression_lib=self.expression_lib)
        end_time = time.time_ns()
        # Update command persistence
        await self.step.workflow.context.database.update_command(command_id, {
            "status": Status.COMPLETED.value,
            "output": str(result),
            "start_time": start_time,
            "end_time": end_time})
        # Return result
        return CWLCommandOutput(value=result, status=Status.COMPLETED, exit_code=0)


class CWLStepCommand(CWLBaseCommand):

    def __init__(self,
                 step: Step,
                 absolute_initial_workdir_allowed: bool = False,
                 expression_lib: Optional[MutableSequence[str]] = None,
                 full_js: bool = False,
                 initial_work_dir: Optional[Union[str, MutableSequence[Any]]] = None,
                 inplace_update: bool = False,
                 time_limit: Optional[Union[int, str]] = None):
        super().__init__(
            step=step,
            absolute_initial_workdir_allowed=absolute_initial_workdir_allowed,
            expression_lib=expression_lib,
            initial_work_dir=initial_work_dir,
            inplace_update=inplace_update,
            full_js=full_js,
            time_limit=time_limit)
        self.input_expressions: MutableMapping[str, str] = {}

    async def execute(self, job: Job) -> CommandOutput:
        context = utils.build_context(
            inputs=job.inputs,
            output_directory=job.output_directory,
            tmp_directory=job.tmp_directory,
            hardware=self.step.workflow.context.scheduler.get_hardware(job.name))
        logger.info('Executing job {job}'.format(job=job.name))
        # Process expressions
        processed_inputs = {}
        for k, v in self.input_expressions.items():
            context = {**context, **{'self': context['inputs'].get(k)}}
            processed_inputs[k] = utils.eval_expression(
                expression=v,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib)
        context['inputs'] = {**context['inputs'], **processed_inputs}
        return CWLCommandOutput(
            value=context['inputs'],
            status=Status.COMPLETED,
            exit_code=0)
